"""
This module contains any class or function related with PELE's Impact
template.
"""

from copy import deepcopy

from simtk import unit

from peleffy.topology import Atom, Bond, Angle, Proper, Improper
from peleffy.topology import ZMatrix


class Impact(object):
    """
    It is in charge of writing a Molecule object as a PELE's Impact
    template.
    """

    def __init__(self, topology):
        """
        Initializes an Impact object.

        Parameters
        ----------
        topology : a Topology object
            The molecular topology representation to write as a
            Impact template

        Examples
        --------

        Write the Impact template of a peleffy's molecule

        >>> from peleffy.topology import Molecule

        >>> molecule = Molecule('molecule.pdb')

        >>> from peleffy.forcefield import OpenForceField

        >>> openff = OpenForceField('openff_unconstrained-1.2.1.offxml')
        >>> parameters = openff.parameterize(molecule)

        >>> from peleffy.topology import Topology

        >>> topology = Topology(molecule, parameters)

        >>> from peleffy.template import Impact

        >>> impact = Impact(topology)
        >>> impact.to_file('molz')

        """
        import peleffy

        # Check input parameters
        if (not isinstance(topology, peleffy.topology.Topology)
                and not
                isinstance(topology, peleffy.topology.topology.Topology)):
            raise TypeError('Invalid input molecule for Impact template')

        self._topology = deepcopy(topology)
        self._molecule = self._topology.molecule
        self._sort()

    def _get_absolute_parent_atom(self):
        """
        It returns the absolute parent atom of the Topology.

        Returns
        -------
        absolute_parent : a peleffy.topology.molecule.Atom
            The absolute parent atom corresponding to the Topology
        """
        for atom in self.topology.atoms:
            if atom.parent is None:
                return atom

        raise Exception('Topology has no absolut parent')

    def _get_core_atoms(self):
        """
        It returns all core atoms of the Topology.

        Returns
        -------
        core_atoms : list[Atom]
            The list of core atoms corresponding to the Topology
        """
        core_atoms = list()

        for atom in self.topology.atoms:
            if atom.core:
                core_atoms.append(atom)

        return core_atoms

    def _get_all_childs_of_atom(self, parent, child_location):
        """
        It returns all child atoms of the supplied Atom, if any.

        Parameters
        ----------
        parent : a peleffy.topology.molecule.Atom
            The Atom whose childs are requested
        child_location : str
            One of ['core', 'side chain']. Whether to look for childs
            in the core or the side chain

        Returns
        -------
        childs : list[Atom]
            The list of childs corresponding to the supplied Atom
        """
        childs = list()

        assert child_location in ['core', 'side chain'], \
            'Unexpected supplied child location'

        for atom in self.topology.atoms:
            if atom.parent == parent:
                if child_location == 'core' and atom.core:
                    childs.append(atom)
                elif child_location == 'side chain' and not atom.core:
                    childs.append(atom)

        return childs

    def _sort(self):
        """Sort and reindex atoms in a Molecule."""

        # Do not sort an empty array of atoms
        if len(self.topology.atoms) == 0:
            return

        sorted_atoms = list()

        # Sort core atoms by parent index
        absolute_parent = self._get_absolute_parent_atom()
        sorted_atoms.append(absolute_parent)
        added_new_child = True

        while added_new_child:
            added_new_child = False

            for atom in sorted_atoms:
                childs = self._get_all_childs_of_atom(
                    atom, child_location='core')

                for child in childs:
                    if child not in sorted_atoms:
                        sorted_atoms.append(child)
                        added_new_child = True

        # Sort non-core atoms by parent index
        added_new_child = True

        while added_new_child:
            added_new_child = False

            for atom in sorted_atoms:
                childs = self._get_all_childs_of_atom(
                    atom, child_location='side chain')

                for child in childs:
                    if child not in sorted_atoms:
                        sorted_atoms.append(child)
                        added_new_child = True

        # Define reindexer and reindex atoms
        reindexer = dict()
        for new_index, atom in enumerate(sorted_atoms):
            old_index = atom._index
            reindexer[old_index] = new_index
            atom.set_index(new_index)

        # Replace old atom list by the sorted one
        self.topology._atoms = sorted_atoms

        # Reindex bonds, angles, propers and impropers
        for bond in self.topology.bonds:
            bond.set_atom1_idx(reindexer[bond.atom1_idx])
            bond.set_atom2_idx(reindexer[bond.atom2_idx])
        for angle in self.topology.angles:
            angle.set_atom1_idx(reindexer[angle.atom1_idx])
            angle.set_atom2_idx(reindexer[angle.atom2_idx])
            angle.set_atom3_idx(reindexer[angle.atom3_idx])
        for proper in self.topology.propers:
            proper.set_atom1_idx(reindexer[proper.atom1_idx])
            proper.set_atom2_idx(reindexer[proper.atom2_idx])
            proper.set_atom3_idx(reindexer[proper.atom3_idx])
            proper.set_atom4_idx(reindexer[proper.atom4_idx])
        for improper in self.topology.impropers:
            improper.set_atom1_idx(reindexer[improper.atom1_idx])
            improper.set_atom2_idx(reindexer[improper.atom2_idx])
            improper.set_atom3_idx(reindexer[improper.atom3_idx])
            improper.set_atom4_idx(reindexer[improper.atom4_idx])

    def to_file(self, path):
        """
        It writes the Impact template to a file.

        Parameters
        ----------
        path : str
            Path to write to
        """
        with open(path, 'w') as file:
            self._write_header(file)
            self._write_resx(file)
            self._write_nbon(file)
            self._write_bond(file)
            self._write_thet(file)
            self._write_phi(file)
            self._write_iphi(file)
            self._write_end(file)

    def _write_header(self, file):
        """
        It writes the header of the Impact file.

        Parameters
        ----------
        file : file object
            File to write to
        """
        import peleffy

        file.write('* LIGAND DATABASE FILE')
        file.write(' ({})'.format(self.topology.parameters.forcefield_name))
        file.write('\n')
        file.write('* File generated with peleffy-{}\n'.format(
            peleffy.__version__))
        file.write('*\n')

    def _write_resx(self, file):
        """
        It writes the resx section of the Impact file.

        Parameters
        ----------
        file : file object
            File to write to
        """
        # template name
        file.write('{:5}'.format(self.molecule.tag))
        # number of non bonding parameters
        file.write('{:6d}'.format(len(self.topology.atoms)))
        # number of bond parameters
        file.write('{:6d}'.format(len(self.topology.bonds)))
        # number of angle parameters
        file.write('{:6d}'.format(len(self.topology.angles)))
        # number of dihedral parameters
        # TODO doublecheck that it is indeed the sum of propers and impropers
        file.write('{:8d}'.format(len(self.topology.propers)
                                  + len(self.topology.impropers)))
        # # number of non-null elements in the interactions matrix
        # TODO It might not be always 0
        file.write('{:8d}'.format(0))
        file.write('\n')

        zmatrix = ZMatrix(self.topology)

        for i, atom in enumerate(self.topology.atoms):
            w_atom = WritableAtom(atom)
            # atom id number
            file.write('{:5d}'.format(w_atom.index))
            file.write(' ')
            file.write('{:5d}'.format(w_atom.parent.index))
            file.write(' ')
            file.write('{:1}'.format(w_atom.core))
            file.write('  ')
            file.write('{:4}'.format(w_atom.OPLS_type))
            file.write('  ')
            file.write('{:4}'.format(w_atom.PDB_name))
            file.write(' ')
            file.write('{:5}'.format(w_atom.unknown))
            file.write(' ')
            file.write('{: 11.6f}'.format(zmatrix[i][0]))
            file.write(' ')
            file.write('{: 11.6f}'.format(zmatrix[i][1]))
            file.write(' ')
            file.write('{: 11.6f}'.format(zmatrix[i][2]))
            file.write('\n')
        # TODO Should we add the interactions matrix here?

    def _write_nbon(self, file):
        """
        It writes the nbon section of the Impact file.

        Parameters
        ----------
        file : file object
            File to write to
        """
        file.write('NBON\n')
        for atom in self.topology.atoms:
            w_atom = WritableAtom(atom)
            # TODO an extra space is found in the IMPACT file generated by
            # PlopRotTemp, consider removing it
            file.write(' ')
            # Atom id
            file.write('{:5d}'.format(w_atom.index))
            file.write(' ')
            # Sigma
            file.write('{: 8.4f}'.format(w_atom.sigma))
            file.write(' ')
            # Epsilon
            file.write('{: 8.4f}'.format(w_atom.epsilon))
            file.write(' ')
            # Charge
            file.write('{: 10.6f}'.format(w_atom.charge))
            file.write(' ')
            # Rad. Non Polar SGB
            file.write('{: 8.4f}'.format(w_atom.born_radius))
            file.write(' ')
            # Rad. Non Polar Type
            file.write('{: 8.4f}'.format(w_atom.SASA_radius))
            file.write(' ')
            # SGB Non Polar gamma
            file.write('{: 13.9f}'.format(w_atom.nonpolar_gamma))
            file.write(' ')
            # SGB Non Polar type
            file.write('{: 13.9f}'.format(w_atom.nonpolar_alpha))
            file.write('\n')

    def _write_bond(self, file):
        """
        It writes the bond section of the Impact file.

        Parameters
        ----------
        file : file object
            File to write to
        """
        file.write('BOND\n')
        for bond in self.topology.bonds:
            w_bond = WritableBond(bond)
            idx1, idx2, spring, eq_dist = [attr[1] for attr in list(w_bond)]
            # TODO an extra space is found in the IMPACT file generated by
            # PlopRotTemp, consider removing it
            file.write(' ')
            # Atom 1 id
            file.write('{:5d}'.format(idx1))
            file.write(' ')
            # Atom 2 id
            file.write('{:5d}'.format(idx2))
            file.write(' ')
            # Spring constant
            file.write('{: 9.3f}'.format(spring))
            file.write(' ')
            # Equilibrium distance
            file.write('{: 6.3f}\n'.format(eq_dist))

    def _write_thet(self, file):
        """
        It writes the thet section of the Impact file.

        Parameters
        ----------
        file : file object
            File to write to
        """
        file.write('THET\n')
        for angle in self.topology.angles:
            w_angle = WritableAngle(angle)
            idx1, idx2, idx3, spring, eq_angl = [attr[1] for attr in
                                                 list(w_angle)]
            # TODO an extra space is found in the IMPACT file generated by
            # PlopRotTemp, consider removing it
            file.write(' ')
            # Atom 1 id
            file.write('{:5d}'.format(idx1))
            file.write(' ')
            # Atom 2 id
            file.write('{:5d}'.format(idx2))
            file.write(' ')
            # Atom 3 id
            file.write('{:5d}'.format(idx3))
            file.write(' ')
            # Spring constant
            file.write('{: 11.5f}'.format(spring))
            # Equilibrium angle
            file.write('{: 11.5f}\n'.format(eq_angl))

    def _write_phi(self, file):
        """
        It writes the phi section of the Impact file.

        Parameters
        ----------
        file : file object
            File to write to
        """
        file.write('PHI\n')
        for proper in self.topology.propers:
            w_proper = WritableProper(proper)
            idx1, idx2, idx3, idx4, constant, prefactor, term, phase = \
                [attr[1] for attr in list(w_proper)]
            # Atom 1 id
            file.write('{:5d}'.format(idx1))
            file.write(' ')
            # Atom 2 id
            file.write('{:5d}'.format(idx2))
            file.write(' ')
            # Atom 3 id
            file.write('{:5d}'.format(idx3))
            file.write(' ')
            # Atom 4 id
            file.write('{:5d}'.format(idx4))
            file.write(' ')
            # Constant
            file.write('{: 9.5f}'.format(constant))
            file.write(' ')
            # Prefactor
            file.write('{: 4.1f}'.format(prefactor))
            file.write(' ')
            # Number of term
            file.write('{:3.1f}'.format(term))
            # Phase (only if different from 0)
            if phase != 0.0:
                file.write(' ')
                file.write('{:5.1f}'.format(phase))
            file.write('\n')

    def _write_iphi(self, file):
        """
        It writes the iphi section of the Impact file.

        Parameters
        ----------
        file : file object
            File to write to
        """
        file.write('IPHI\n')
        for improper in self.topology.impropers:
            w_improper = WritableImproper(improper)
            idx1, idx2, idx3, idx4, constant, prefactor, term = \
                [attr[1] for attr in list(w_improper)]
            # TODO an extra space is found in the IMPACT file generated by
            # PlopRotTemp, consider removing it
            file.write(' ')
            # Atom 1 id
            file.write('{:5d}'.format(idx1))
            file.write(' ')
            # Atom 2 id
            file.write('{:5d}'.format(idx2))
            file.write(' ')
            # Atom 3 id
            file.write('{:5d}'.format(idx3))
            file.write(' ')
            # Atom 4 id
            file.write('{:5d}'.format(idx4))
            file.write(' ')
            # Constant
            file.write('{: 9.5f}'.format(constant))
            file.write(' ')
            # Prefactor
            file.write('{: 4.1f}'.format(prefactor))
            file.write(' ')
            # Number of term
            file.write('{:3.1f}'.format(term))
            file.write('\n')

    def _write_end(self, file):
        """
        It writes the ending line of the Impact file.

        Parameters
        ----------
        file : file object
            File to write to
        """
        file.write('END\n')

    @property
    def molecule(self):
        """
        The peleffy's Molecule.

        Returns
        -------
        molecule : a peleffy.topology.Molecule
            The peleffy's Molecule object
        """
        return self._molecule

    @property
    def topology(self):
        """
        The peleffy's Topology.

        Returns
        -------
        topology : a peleffy.topology.Topology
            The peleffy's Topology object
        """
        return self._topology


class WritableWrapper(object):
    """
    Wrapper class for writable parameters.
    """

    @staticmethod
    def none_to_zero(f):
        """
        It converts a returned None to zero.

        Parameters
        ----------
        f : function
            The function to apply the decorator to

        Returns
        -------
        out : function's output
            It is set to zero in case that it is None
        """
        def function_wrapper(*args, **kwargs):
            out = f(*args, **kwargs)
            if out is None:
                out = int(0)
            return out
        return function_wrapper

    @staticmethod
    def dummy_to_writable(f):
        """
        It converts a returned DummyAtom to a WritableAtom.

        Parameters
        ----------
        f : function
            The function to apply the decorator to

        Returns
        -------
        out : WritableAtom
            A WritableAtom object
        """
        def function_wrapper(*args, **kwargs):
            out = f(*args, **kwargs)
            out = WritableAtom(out)
            return out
        return function_wrapper

    @staticmethod
    def none_to_dummy(f):
        """
        It converts a returned None to a DummyAtom.

        Parameters
        ----------
        f : function
            The function to apply the decorator to

        Returns
        -------
        out : DummyAtom
            A DummyAtom object
        """
        import peleffy

        def function_wrapper(*args, **kwargs):
            out = f(*args, **kwargs)
            if out is None:
                out = peleffy.topology.elements.DummyAtom(index=-1)
            return out
        return function_wrapper

    @staticmethod
    def in_angstrom(f):
        """
        It expresses a simtk.unit.Quantity in angstroms.

        Parameters
        ----------
        f : function
            The function to apply the decorator to

        Returns
        -------
        out : float
            simtk.unit.Quantity expressed in angstroms
        """
        def function_wrapper(*args, **kwargs):
            out = f(*args, **kwargs)
            return out.value_in_unit(unit.angstrom)
        return function_wrapper

    @staticmethod
    def in_kcalmol(f):
        """
        It expresses a simtk.unit.Quantity in kcal/mol.

        Parameters
        ----------
        f : function
            The function to apply the decorator to

        Returns
        -------
        out : float
            simtk.unit.Quantity expressed in kcal/mol
        """
        def function_wrapper(*args, **kwargs):
            out = f(*args, **kwargs)
            return out.value_in_unit(unit.kilocalorie / unit.mole)
        return function_wrapper

    @staticmethod
    def in_elementarycharge(f):
        """
        It expresses a simtk.unit.Quantity in elementary charges.

        Parameters
        ----------
        f : function
            The function to apply the decorator to

        Returns
        -------
        out : float
            simtk.unit.Quantity expressed in elementary charges
        """
        def function_wrapper(*args, **kwargs):
            out = f(*args, **kwargs)
            return out.value_in_unit(unit.elementary_charge)
        return function_wrapper

    @staticmethod
    def in_kcal_rad2mol(f):
        """
        It expresses a simtk.unit.Quantity in kcal/rad2mol.

        Parameters
        ----------
        f : function
            The function to apply the decorator to

        Returns
        -------
        out : float
            simtk.unit.Quantity expressed in kcal/rad2mol
        """
        def function_wrapper(*args, **kwargs):
            out = f(*args, **kwargs)
            return out.value_in_unit(unit.kilocalorie
                                     / (unit.radian**2 * unit.mole))
        return function_wrapper

    @staticmethod
    def in_deg(f):
        """
        It expresses a simtk.unit.Quantity in degrees.

        Parameters
        ----------
        f : function
            The function to apply the decorator to

        Returns
        -------
        out : float
            simtk.unit.Quantity expressed in degrees
        """
        def function_wrapper(*args, **kwargs):
            out = f(*args, **kwargs)
            return out.value_in_unit(unit.degree)
        return function_wrapper

    @staticmethod
    def in_kcal_angstrom2mol(f):
        """
        It expresses a simtk.unit.Quantity in kcal/angstrom2mol.

        Parameters
        ----------
        f : function
            The function to apply the decorator to

        Returns
        -------
        out : float
            simtk.unit.Quantity expressed in kcal/angstrom2mol
        """
        def function_wrapper(*args, **kwargs):
            out = f(*args, **kwargs)
            return out.value_in_unit(unit.kilocalorie
                                     / (unit.angstrom**2 * unit.mole))
        return function_wrapper


class WritableAtom(Atom, WritableWrapper):
    """
    Writable peleffy's Atom class
    """

    def __init__(self, atom):
        """
        It initializes a WritableAtom object.

        Parameters
        ----------
        atom : a peleffy.topology.molecule.Atom
            The Atom to create the WritableAtom with
        """
        import peleffy

        # We do not want to modify the original object
        atom = deepcopy(atom)

        assert isinstance(atom, (peleffy.topology.Atom,
                                 peleffy.topology.elements.Atom,
                                 peleffy.topology.elements.DummyAtom)), \
            'Wrong type: {}'.format(type(atom))

        super().__init__(atom.index, atom.core, atom.OPLS_type, atom.PDB_name,
                         atom.unknown, atom.x, atom.y, atom.z, atom.sigma,
                         atom.epsilon, atom.charge, atom.born_radius,
                         atom.SASA_radius, atom.nonpolar_gamma,
                         atom.nonpolar_alpha, atom.parent)

    @property
    @WritableWrapper.dummy_to_writable
    @WritableWrapper.none_to_dummy
    def parent(self):
        """
        Atom's parent.

        Returns
        -------
        parent : a peleffy.topology.molecule.Atom
            The parent of this Atom object
        """
        return super().parent

    @property
    def index(self):
        """
        Atom's index.

        Returns
        -------
        index : int
            The index of this Atom object
        """
        return int(self._index) + 1

    @property
    def core(self):
        """
        Atom's core char.

        Returns
        -------
        index : str
            The core type of this Atom object
        """
        if self._core:
            return 'M'
        else:
            return 'S'

    @property
    def OPLS_type(self):
        """
        Atom's OPLS type.

        .. todo ::

           * Consider removing any reference to OPLS, if possible
             Otherwise, use SMIRks to find the best match

        Returns
        -------
        index : str
            The OLPS type of this Atom object
        """
        return super().OPLS_type  # stands for OpenForceField type

    # TODO
    @property
    @WritableWrapper.none_to_zero
    def unknown(self):
        """
        Atom's unknown int.

        .. todo ::

           * Review the actual purpose of this attribute in PELE

        Returns
        -------
        unknown : int
            The unknown int of this Atom object
        """
        return super().unknown

    @property
    @WritableWrapper.in_angstrom
    def sigma(self):
        """
        Atom's sigma.

        Returns
        -------
        sigma : float
            The sigma of this Atom object, expressed in angstroms
        """
        return super().sigma

    @property
    @WritableWrapper.in_kcalmol
    def epsilon(self):
        """
        Atom's epsilon.

        Returns
        -------
        epsilon : float
            The epsilon of this Atom object, expressed in kcal/mol
        """
        return super().epsilon

    @property
    @WritableWrapper.in_elementarycharge
    def charge(self):
        """
        Atom's charge.

        Returns
        -------
        charge : float
            The charge of this Atom object, expressed in elementary units
        """
        return super().charge

    @property
    @WritableWrapper.in_angstrom
    def born_radius(self):
        """
        Atom's Born radius.

        Returns
        -------
        born_radius : float
            The Born radius of this Atom object
        """
        if super().born_radius is None:
            return unit.Quantity(0, unit.angstroms)

        return super().born_radius

    @property
    @WritableWrapper.in_angstrom
    def SASA_radius(self):
        """
        Atom's SASA radius.

        Returns
        -------
        SASA_radius : float
            The SASA radius of this Atom object, expressed in angstroms
        """
        return super().SASA_radius

    @property
    @WritableWrapper.none_to_zero
    def nonpolar_gamma(self):
        """
        Atom's nonpolar gamma.

        Returns
        -------
        nonpolar_gamma : float
            The nonpolar gamma of this Atom object
        """
        return super().nonpolar_gamma

    @property
    @WritableWrapper.none_to_zero
    def nonpolar_alpha(self):
        """
        Atom's nonpolar alpha.

        Returns
        -------
        nonpolar_alpha : float
            The nonpolar alpha of this Atom object
        """
        return super().nonpolar_alpha


class WritableBond(Bond, WritableWrapper):
    """
    Writable peleffy's Bond class
    """

    def __init__(self, bond):
        """
        It initializes a WritableBond object.

        Parameters
        ----------
        bond : a peleffy.topology.Bond
            The Bond to create the WritableBond with
        """
        import peleffy

        # We do not want to modify the original object
        bond = deepcopy(bond)

        assert isinstance(bond, (peleffy.topology.Bond,
                                 peleffy.topology.elements.Bond)), \
            'Wrong type: {}'.format(type(bond))

        super().__init__(index=bond.index, atom1_idx=bond.atom1_idx,
                         atom2_idx=bond.atom2_idx,
                         spring_constant=bond.spring_constant,
                         eq_dist=bond.eq_dist)

    @property
    def atom1_idx(self):
        """
        Bond's atom1 index.

        Returns
        -------
        atom1_idx : int
            The index of the first atom involved in this Bond object
        """
        return super().atom1_idx + 1

    @property
    def atom2_idx(self):
        """
        Bond's atom2 index.

        Returns
        -------
        atom2_idx : int
            The index of the second atom involved in this Bond object
        """
        return super().atom2_idx + 1

    @property
    @WritableWrapper.in_kcal_angstrom2mol
    def spring_constant(self):
        """
        Bond's spring constant.

        Returns
        -------
        spring_constant : float
            The spring constant of this Bond object
        """
        return super().spring_constant

    @property
    @WritableWrapper.in_angstrom
    def eq_dist(self):
        """
        Bond's equilibrium distance.

        Returns
        -------
        eq_dist : float
            The equilibrium distance of this Bond object
        """
        return super().eq_dist


class WritableAngle(Angle, WritableWrapper):
    """
    Writable peleffy's Angle class
    """

    def __init__(self, angle):
        """
        It initializes a WritableAngle object.

        Parameters
        ----------
        angle : a peleffy.topology.Angle
            The Angle to create the WritableAngle with
        """
        import peleffy

        # We do not want to modify the original object
        angle = deepcopy(angle)

        assert isinstance(angle, (peleffy.topology.Angle,
                                  peleffy.topology.elements.Angle)), \
            'Wrong type: {}'.format(type(angle))

        super().__init__(index=angle.index, atom1_idx=angle.atom1_idx,
                         atom2_idx=angle.atom2_idx, atom3_idx=angle.atom3_idx,
                         spring_constant=angle.spring_constant,
                         eq_angle=angle.eq_angle)

    @property
    def atom1_idx(self):
        """
        Angle's atom1 index.

        Returns
        -------
        atom1_idx : int
            The index of the first atom involved in this Angle object
        """
        return super().atom1_idx + 1

    @property
    def atom2_idx(self):
        """
        Angle's atom2 index.

        Returns
        -------
        atom2_idx : int
            The index of the second atom involved in this Angle object
        """
        return super().atom2_idx + 1

    @property
    def atom3_idx(self):
        """
        Angle's atom3 index.

        Returns
        -------
        atom3_idx : int
            The index of the third atom involved in this Angle object
        """
        return super().atom3_idx + 1

    @property
    @WritableWrapper.in_kcal_rad2mol
    def spring_constant(self):
        """
        Angle's spring constant.

        Returns
        -------
        spring_constant : float
            The spring constant of this Angle object
        """
        return super().spring_constant

    @property
    @WritableWrapper.in_deg
    def eq_angle(self):
        """
        Angle's equilibrium distance.

        Returns
        -------
        eq_angle : float
            The equilibrium angle of this Angle object
        """
        return super().eq_angle


class WritableProper(Proper, WritableWrapper):
    """
    Writable peleffy's Proper class
    """

    def __init__(self, proper):
        """
        It initializes a WritableProper object.

        Parameters
        ----------
        proper : a peleffy.topology.Proper
            The Proper to create the WritableProper with
        """
        import peleffy

        # We do not want to modify the original object
        proper = deepcopy(proper)

        assert isinstance(proper, (peleffy.topology.Proper,
                                   peleffy.topology.elements.Proper)), \
            'Wrong type: {}'.format(type(proper))

        super().__init__(index=proper.index, atom1_idx=proper.atom1_idx,
                         atom2_idx=proper.atom2_idx,
                         atom3_idx=proper.atom3_idx,
                         atom4_idx=proper.atom4_idx,
                         periodicity=proper.periodicity,
                         prefactor=proper.prefactor,
                         constant=proper.constant,
                         phase=proper.phase)

        self.exclude = proper.exclude

    @property
    def atom1_idx(self):
        """
        Proper's atom1 index.

        Returns
        -------
        atom1_idx : int
            The index of the first atom involved in this Proper object
        """
        return super().atom1_idx + 1

    @property
    def atom2_idx(self):
        """
        Proper's atom2 index.

        Returns
        -------
        atom2_idx : int
            The index of the second atom involved in this Proper object
        """
        return super().atom2_idx + 1

    @property
    def atom3_idx(self):
        """
        Proper's atom3 index.

        Returns
        -------
        atom3_idx : int
            The index of the third atom involved in this Proper object
        """
        if self.exclude:
            return (super().atom3_idx + 1) * -1
        else:
            return super().atom3_idx + 1

    @property
    def atom4_idx(self):
        """
        Proper's atom4 index.

        Returns
        -------
        atom4_idx : int
            The index of the fourth atom involved in this Proper object
        """
        return super().atom4_idx + 1

    @property
    @WritableWrapper.in_kcalmol
    def constant(self):
        """
        Proper's constant.

        Returns
        -------
        constant : float
            The constant of this Proper object, expressed in kcal/mol
        """
        return super().constant

    @property
    @WritableWrapper.in_deg
    def phase(self):
        """
        Proper's phase constant.

        Returns
        -------
        phase : float
            The phase constant of this Proper object, expressed in
            degrees
        """
        return super().phase


class WritableImproper(Improper, WritableWrapper):
    """
    Writable peleffy's Improper class
    """

    def __init__(self, improper):
        """
        It initializes a WritableImproper object.

        Parameters
        ----------
        improper : a peleffy.topology.Improper
            The Improper to create the WritableImproper with
        """
        import peleffy

        # We do not want to modify the original object
        improper = deepcopy(improper)

        assert isinstance(improper, (peleffy.topology.Improper,
                                     peleffy.topology.elements.Improper)), \
            'Wrong type: {}'.format(type(improper))

        super().__init__(index=improper.index, atom1_idx=improper.atom1_idx,
                         atom2_idx=improper.atom2_idx,
                         atom3_idx=improper.atom3_idx,
                         atom4_idx=improper.atom4_idx,
                         periodicity=improper.periodicity,
                         prefactor=improper.prefactor,
                         constant=improper.constant)

    @property
    def atom1_idx(self):
        """
        Improper's atom1 index.

        Returns
        -------
        atom1_idx : int
            The index of the first atom involved in this Improper object
        """
        return super().atom1_idx + 1

    @property
    def atom2_idx(self):
        """
        Improper's atom2 index.

        Returns
        -------
        atom2_idx : int
            The index of the second atom involved in this Improper object
        """
        return super().atom2_idx + 1

    @property
    def atom3_idx(self):
        """
        Improper's atom3 index.

        Returns
        -------
        atom3_idx : int
            The index of the third atom involved in this Improper object
        """
        return super().atom3_idx + 1

    @property
    def atom4_idx(self):
        """
        Improper's atom4 index.

        Returns
        -------
        atom4_idx : int
            The index of the fourth atom involved in this Improper object
        """
        return super().atom4_idx + 1

    @property
    @WritableWrapper.in_kcalmol
    def constant(self):
        """
        Improper's constant.

        Returns
        -------
        constant : float
            The constant of this Improper object, expressed in kcal/mol
        """
        return super().constant



"""
This module contains the classes and functions to generate the zmatrix of
a molecule.
"""

from copy import deepcopy

import numpy as np

from peleffy.topology.elements import DummyAtom


class ZMatrix(np.ndarray):
    """
    It generates the zmatrix of a molecule as a numpy.array.

    Inspired by the PlopRotTemp algorithm.
    """

    def __init__(self, topology):
        """
        It initializes a ZMatrix object.

        Parameters
        ----------
        topology : a peleffy.topology.Topology
            The molecular topology representation to generate the
            zmatrix with

        Examples
        --------

        Given a molecular topology, build its Z-matrix

        >>> from peleffy.topology import Molecule

        >>> molecule = Molecule(smiles='Cc1ccccc1')

        >>> from peleffy.forcefield import OpenForceField

        >>> openff = OpenForceField('openff_unconstrained-1.2.1.offxml')
        >>> parameters = openff.parameterize(molecule)

        >>> from peleffy.topology import Topology
        >>> topology = Topology(molecule, parameters)

        >>> from peleffy.topology import ZMatrix

        >>> zmatrix = ZMatrix(topology)

        """
        # We will work on a copy of the topology's object to modify it freely
        self._topology = deepcopy(topology)

    def __new__(cls, topology):
        """
        It customizes the creation of the numpy.array.

        Parameters
        ----------
        topology : a peleffy.topology.Topology
            The molecular topology representation to generate the
            zmatrix with
        """
        topology = deepcopy(topology)
        obj = np.zeros((len(topology.atoms), 3)).view(cls)
        coords = cls._extract_coords(topology)
        obj = cls._build_zmatrix(cls, obj, coords, topology)

        return obj

    @staticmethod
    def _extract_coords(topology):
        """
        It extracts the coordinates of the molecule's atoms.

        Parameters
        ----------
        topology : a peleffy.topology.Topology
            The molecular topology representation to generate the
            zmatrix with

        Returns
        -------
        coords : list[tuple[float]]
            The coordinates of the molecule
        """
        coords = list()
        for atom in topology.atoms:
            coords.append((atom.x, atom.y, atom.z))

        return coords

    @staticmethod
    def _get_absolute_parent(topology):
        """
        It returns the absolute parent in the topology of the molecule.

        Parameters
        ----------
        topology : a peleffy.topology.Topology
            The molecular topology representation to generate the
            zmatrix with

        Returns
        -------
        absolute_parent : peleffy.topology.molecule.Atom
            The absolute parent of the molecule
        """
        absolute_parent = list()
        for atom in topology.atoms:
            if atom.parent is None:
                absolute_parent.append(atom)

        assert len(absolute_parent) == 1, 'Only 1 absolute parent is expected'

        return absolute_parent[0]

    @staticmethod
    def _calculate_bond(x1, y1, z1, x2, y2, z2):
        """
        It calculates the bond distance between two sets of coordinates.

        Parameters
        ----------
        x1 : float
            The x1 coordinate
        y1 : float
            The y1 coordinate
        z1 : float
            The z1 coordinate
        x2 : float
            The x2 coordinate
        y2 : float
            The y2 coordinate
        z2 : float
            The z2 coordinate

        Returns
        -------
        distance : float
            The bond distance between the two sets of coordinates
        """
        dx = x1 - x2
        dy = y1 - y2
        dz = z1 - z2

        return np.sqrt(dx * dx + dy * dy + dz * dz)

    @staticmethod
    def _calculate_angle(x1, y1, z1, x2, y2, z2, x3, y3, z3):
        """
        It calculates the angle between three sets of coordinates.

        Parameters
        ----------
        x1 : float
            The x1 coordinate
        y1 : float
            The y1 coordinate
        z1 : float
            The z1 coordinate
        x2 : float
            The x2 coordinate
        y2 : float
            The y2 coordinate
        z2 : float
            The z2 coordinate
        x3 : float
            The x3 coordinate
        y3 : float
            The y3 coordinate
        z3 : float
            The z3 coordinate

        Returns
        -------
        angle : float
            The angle between the three sets of coordinates
        """
        dx_12 = x1 - x2
        dy_12 = y1 - y2
        dz_12 = z1 - z2

        dx_31 = x3 - x1
        dy_31 = y3 - y1
        dz_31 = z3 - z1

        vdot = dx_12 * dx_31 + dy_12 * dy_31 + dz_12 * dz_31

        assert np.fabs(vdot) > 0, 'A non-zero angle is expected'

        d12 = np.sqrt(dx_12 * dx_12 + dy_12 * dy_12 + dz_12 * dz_12)
        d31 = np.sqrt(dx_31 * dx_31 + dy_31 * dy_31 + dz_31 * dz_31)

        xang = vdot / (d12 * d31)

        if xang - 1.0 > -0.0000000001:
            return 0.0
        elif xang + 1.0 < 0.0000000001:
            return np.pi

        return np.arccos(xang) * 180.0 / np.pi

    @staticmethod
    def _calculate_dihedral(x1, y1, z1, x2, y2, z2, x3, y3, z3, x4, y4, z4):
        """
        It calculates the dihedral between four sets of coordinates.

        Parameters
        ----------
        x1 : float
            The x1 coordinate
        y1 : float
            The y1 coordinate
        z1 : float
            The z1 coordinate
        x2 : float
            The x2 coordinate
        y2 : float
            The y2 coordinate
        z2 : float
            The z2 coordinate
        x3 : float
            The x3 coordinate
        y3 : float
            The y3 coordinate
        z3 : float
            The z3 coordinate
        x4 : float
            The x4 coordinate
        y4 : float
            The y4 coordinate
        z4 : float
            The z4 coordinate

        Returns
        -------
        dihedral : float
            The dihedral between the four sets of coordinates
        """
        dx_12 = x1 - x2
        dy_12 = y1 - y2
        dz_12 = z1 - z2

        dx_32 = x3 - x2
        dy_32 = y3 - y2
        dz_32 = z3 - z2

        dx_34 = x3 - x4
        dy_34 = y3 - y4
        dz_34 = z3 - z4

        ax = dy_12 * dz_32 - dz_12 * dy_32
        ay = dz_12 * dx_32 - dx_12 * dz_32
        az = dx_12 * dy_32 - dy_12 * dx_32
        cx = dy_32 * dz_34 - dz_32 * dy_34
        cy = dz_32 * dx_34 - dx_32 * dz_34
        cz = dx_32 * dy_34 - dy_32 * dx_34

        rac = ax * cx + ay * cy + az * cz
        ra = ax * ax + ay * ay + az * az
        rc = cx * cx + cy * cy + cz * cz

        cosang = rac / np.sqrt(ra * rc)

        if cosang - 1.0 > -0.00000000001:
            phi = 0.0
        elif cosang + 1.0 < 0.00000000001:
            phi = np.pi
        else:
            phi = np.arccos(cosang)

        s = dx_12 * cx + dy_12 * cy + dz_12 * cz
        if (s < 0):
            phi = -phi  # to account for phi between pi and 2pi

        return phi * 180.0 / np.pi

    @staticmethod
    def _build_zmatrix(cls, obj, coords, topology):
        """
        It buils the zmatrix.

        Parameters
        ----------
        cls : ZMatrix class
            The ZMatrix class
        obj : ZMatrix object
            The ZMatrix object
        coords : list[tuple[float]]
            The coordinates of the molecule
        topology : a peleffy.topology.Topology
            The molecular topology representation to generate the
            zmatrix with

        Returns
        -------
        obj :  ZMatrix object
            The ZMatrix object which is a numpy.array with the corresponding
            zmatrix initialized.
        """
        dummy1 = DummyAtom(index=-3, PDB_name='DUM1', parent=None)
        dummy2 = DummyAtom(index=-2, PDB_name='DUM2', parent=dummy1)
        dummy3 = DummyAtom(index=-1, PDB_name='DUM3', parent=dummy2)
        absolute_parent = cls._get_absolute_parent(topology)
        absolute_parent.set_parent(dummy3)

        dummy1.set_coords([0.0, 0.0, 0.0])  # at origin
        dummy2.set_coords([0.0, 0.0, 1.0])  # at z axis
        dummy3.set_coords([1.0, 0.0, 0.0])  # at x axis

        for i, atom in enumerate(topology.atoms):
            atom1 = atom
            atom2 = atom1.parent
            atom3 = atom2.parent
            atom4 = atom3.parent

            assert atom1 is not None and atom2 is not None \
                and atom3 is not None and atom4 is not None, \
                'A None as parent is not expected'

            x1, y1, z1 = (atom1.x, atom1.y, atom1.z)
            x2, y2, z2 = (atom2.x, atom2.y, atom2.z)
            x3, y3, z3 = (atom3.x, atom3.y, atom3.z)
            x4, y4, z4 = (atom4.x, atom4.y, atom4.z)

            obj[i][0] = cls._calculate_bond(x1, y1, z1,
                                            x2, y2, z2)
            obj[i][1] = cls._calculate_angle(x1, y1, z1,
                                             x2, y2, z2,
                                             x3, y3, z3)
            obj[i][2] = cls._calculate_dihedral(x1, y1, z1,
                                                x2, y2, z2,
                                                x3, y3, z3,
                                                x4, y4, z4)

        return obj



"""
This module contains the tests to check the peleffy's toolkits.
"""

import numpy as np
import pytest

from peleffy.forcefield.parameters import OPLS2005ParameterWrapper
from simtk import unit


FORCEFIELD_NAME = 'openff_unconstrained-1.2.0.offxml'
METHANE_OPLS_PARAMETERS = OPLS2005ParameterWrapper({
    'atom_names': [' C1 ', ' H2 ', ' H3 ', ' H4 ', ' H5 '],
    'atom_types': ['CT', 'HC', 'HC', 'HC', 'HC'],
    'charges': [unit.Quantity(-0.24, unit.elementary_charge),
                unit.Quantity(0.06, unit.elementary_charge),
                unit.Quantity(0.06, unit.elementary_charge),
                unit.Quantity(0.06, unit.elementary_charge),
                unit.Quantity(0.06, unit.elementary_charge)],
    'sigmas': [unit.Quantity(3.5, unit.angstrom),
               unit.Quantity(2.5, unit.angstrom),
               unit.Quantity(2.5, unit.angstrom),
               unit.Quantity(2.5, unit.angstrom),
               unit.Quantity(2.5, unit.angstrom)],
    'epsilons': [unit.Quantity(0.066, unit.kilocalorie / unit.mole),
                 unit.Quantity(0.03, unit.kilocalorie / unit.mole),
                 unit.Quantity(0.03, unit.kilocalorie / unit.mole),
                 unit.Quantity(0.03, unit.kilocalorie / unit.mole),
                 unit.Quantity(0.03, unit.kilocalorie / unit.mole)],
    'bonds': [{'atom1_idx': 0, 'atom2_idx': 1,
               'spring_constant': unit.Quantity(
                   340.0, unit.kilocalorie / (unit.angstrom ** 2
                                              * unit.mole)),
               'eq_dist': unit.Quantity(1.09, unit.angstrom)},
              {'atom1_idx': 0, 'atom2_idx': 2,
               'spring_constant': unit.Quantity(
                   340.0, unit.kilocalorie / (unit.angstrom ** 2
                                              * unit.mole)),
               'eq_dist': unit.Quantity(1.09, unit.angstrom)},
              {'atom1_idx': 0,
               'atom2_idx': 3,
               'spring_constant': unit.Quantity(340.0, unit.kilocalorie
                                                / (unit.angstrom ** 2
                                                   * unit.mole)),
               'eq_dist': unit.Quantity(1.09, unit.angstrom)},
              {'atom1_idx': 0, 'atom2_idx': 4,
               'spring_constant': unit.Quantity(340.0, unit.kilocalorie
                                                / (unit.angstrom ** 2
                                                   * unit.mole)),
               'eq_dist': unit.Quantity(1.09, unit.angstrom)}],
    'angles': [{'atom1_idx': 1, 'atom2_idx': 0, 'atom3_idx': 2,
                'spring_constant': unit.Quantity(
                    33.0, unit.kilocalorie / (unit.mole
                                              * unit.radian ** 2)),
                'eq_angle': unit.Quantity(107.8, unit.degree)},
               {'atom1_idx': 1, 'atom2_idx': 0, 'atom3_idx': 3,
                'spring_constant': unit.Quantity(
                    33.0, unit.kilocalorie / (unit.mole
                                              * unit.radian ** 2)),
                'eq_angle': unit.Quantity(107.8, unit.degree)},
               {'atom1_idx': 1, 'atom2_idx': 0, 'atom3_idx': 4,
                'spring_constant': unit.Quantity(
                    33.0, unit.kilocalorie / (unit.mole
                                              * unit.radian ** 2)),
                'eq_angle': unit.Quantity(107.8, unit.degree)},
               {'atom1_idx': 2, 'atom2_idx': 0, 'atom3_idx': 3,
                'spring_constant': unit.Quantity(
                    33.0, unit.kilocalorie / (unit.mole
                                              * unit.radian ** 2)),
                'eq_angle': unit.Quantity(107.8, unit.degree)},
               {'atom1_idx': 2, 'atom2_idx': 0, 'atom3_idx': 4,
                'spring_constant': unit.Quantity(
                    33.0, unit.kilocalorie / (unit.mole
                                              * unit.radian ** 2)),
                'eq_angle': unit.Quantity(107.8, unit.degree)},
               {'atom1_idx': 3, 'atom2_idx': 0, 'atom3_idx': 4,
                'spring_constant': unit.Quantity(
                    33.0, unit.kilocalorie / (unit.mole
                                              * unit.radian ** 2)),
                'eq_angle': unit.Quantity(107.8, unit.degree)}],
    'SGB_radii': [unit.Quantity(1.975, unit.angstrom),
                  unit.Quantity(1.425, unit.angstrom),
                  unit.Quantity(1.425, unit.angstrom),
                  unit.Quantity(1.425, unit.angstrom),
                  unit.Quantity(1.425, unit.angstrom)],
    'vdW_radii': [unit.Quantity(1.75, unit.angstrom),
                  unit.Quantity(1.25, unit.angstrom),
                  unit.Quantity(1.25, unit.angstrom),
                  unit.Quantity(1.25, unit.angstrom),
                  unit.Quantity(1.25, unit.angstrom)],
    'gammas': [0.005, 0.00859824, 0.00859824, 0.00859824, 0.00859824],
    'alphas': [-0.74168571, 0.268726247, 0.268726247, 0.268726247,
               0.268726247]})


class TestSchrodingerToolkitWrapper(object):
    """
    It wraps all tests that check the SchrodingerToolkitWrapperMolecularGraph
    class.
    """

    def test_get_Schrodinger_parameters(self):
        """
        It tests the standard methods to obtain Schrodinger parameters
        from an peleffy's Molecule.
        """
        from peleffy.topology import Molecule
        from peleffy.forcefield import OPLS2005ForceField
        from peleffy.utils.toolkits import ToolkitUnavailableException

        # Load benzene ring
        molecule = Molecule(smiles='c1ccccc1')

        # Load OPLS2005 force field
        opls2005 = OPLS2005ForceField()

        # Ensure SCHRODINGER is not in the environment
        import os
        if 'SCHRODINGER' in os.environ:
            del(os.environ['SCHRODINGER'])

        with pytest.raises(ToolkitUnavailableException):
            opls2005.parameterize(molecule, charge_method='gasteiger')


class TestRDKitToolkitWrapper(object):
    """
    It wraps all tests that check the RDKitToolkitWrapper class.
    """

    def test_conformer_setter(self):
        """It checks the conformer setter of the RDKit toolkit"""
        from peleffy.topology import Molecule
        from rdkit import Chem
        from copy import deepcopy

        from peleffy.utils import get_data_file_path

        # Load molecule
        mol = Molecule(get_data_file_path('ligands/propionic_acid.pdb'))

        # Choose a dihedral to track
        dihedral = (0, 1, 2, 3)

        # Get initial dihedral's theta
        conformer = mol.rdkit_molecule.GetConformer()
        initial_theta = Chem.rdMolTransforms.GetDihedralDeg(conformer,
                                                            *dihedral)

        if initial_theta < -179:
            initial_theta += 180.0
        elif initial_theta > 179:
            initial_theta -= 180.0

        assert abs(initial_theta - -0.002) < 10e-3, \
            'Unexpected initial theta value'

        # Get a copy of the rdkit's molecule representation
        rdkit_mol = deepcopy(mol.rdkit_molecule)

        # Modify its conformer
        conformer = rdkit_mol.GetConformer()
        Chem.rdMolTransforms.SetDihedralDeg(conformer, *dihedral, 90)
        new_theta = Chem.rdMolTransforms.GetDihedralDeg(conformer,
                                                        *dihedral)

        assert abs(new_theta - 89.999) < 10e-3, \
            'Unexpected new theta value'

        # Set new conformer to peleffy molecule
        mol.set_conformer(conformer)

        # Check new set theta value
        conformer = mol.rdkit_molecule.GetConformer()
        new_set_theta = Chem.rdMolTransforms.GetDihedralDeg(conformer,
                                                            *dihedral)

        assert abs(new_set_theta - 89.999) < 10e-3, \
            'Unexpected new set theta value'

    def test_atom_degrees(self):
        """It checks that the atom degree getter works well."""

        from peleffy.topology import Molecule
        from peleffy.utils.toolkits import RDKitToolkitWrapper
        from peleffy.utils import get_data_file_path

        wrapper = RDKitToolkitWrapper()

        pdb_path = get_data_file_path('ligands/methane.pdb')
        m = Molecule(pdb_path)
        degree_by_name = dict(zip(wrapper.get_atom_names(m),
                                  wrapper.get_atom_degrees(m)))

        assert degree_by_name == {' C1 ': 4, ' H1 ': 1, ' H2 ': 1,
                                  ' H3 ': 1, ' H4 ': 1}, \
            'Unexpected pairing between atom names and degrees'

        pdb_path = get_data_file_path('ligands/ethylene.pdb')
        m = Molecule(pdb_path)
        degree_by_name = dict(zip(wrapper.get_atom_names(m),
                                  wrapper.get_atom_degrees(m)))

        assert degree_by_name == {' C1 ': 3, ' C2 ': 3, ' H1 ': 1,
                                  ' H2 ': 1, ' H3 ': 1, ' H4 ': 1}, \
            'Unexpected pairing between atom names and degrees'

        pdb_path = get_data_file_path('ligands/acetylene.pdb')
        m = Molecule(pdb_path)
        degree_by_name = dict(zip(wrapper.get_atom_names(m),
                                  wrapper.get_atom_degrees(m)))

        assert degree_by_name == {' C1 ': 2, ' C2 ': 2, ' H1 ': 1,
                                  ' H2 ': 1}, \
            'Unexpected pairing between atom names and degrees'

        pdb_path = get_data_file_path('ligands/propionic_acid.pdb')
        m = Molecule(pdb_path)
        degree_by_name = dict(zip(wrapper.get_atom_names(m),
                                  wrapper.get_atom_degrees(m)))

        assert degree_by_name == {' C1 ': 4, ' C2 ': 4, ' C3 ': 3,
                                  ' O1 ': 1, ' O2 ': 2, ' H1 ': 1,
                                  ' H2 ': 1, ' H3 ': 1, ' H4 ': 1,
                                  ' H5 ': 1, ' H6 ': 1}, \
            'Unexpected pairing between atom names and degrees'

        pdb_path = get_data_file_path('ligands/trimethylglycine.pdb')
        m = Molecule(pdb_path)
        degree_by_name = dict(zip(wrapper.get_atom_names(m),
                                  wrapper.get_atom_degrees(m)))

        assert degree_by_name == {' C1 ': 4, ' N1 ': 4, ' C2 ': 4,
                                  ' C3 ': 4, ' C4 ': 4, ' C5 ': 3,
                                  ' O1 ': 1, ' O2 ': 1, ' H1 ': 1,
                                  ' H2 ': 1, ' H3 ': 1, ' H4 ': 1,
                                  ' H5 ': 1, ' H6 ': 1, ' H7 ': 1,
                                  ' H8 ': 1, ' H9 ': 1, ' H10': 1,
                                  ' H11': 1}, \
            'Unexpected pairing between atom names and degrees'

        pdb_path = get_data_file_path('ligands/malonate.pdb')
        m = Molecule(pdb_path)
        degree_by_name = dict(zip(wrapper.get_atom_names(m),
                                  wrapper.get_atom_degrees(m)))

        assert degree_by_name == {' O1 ': 1, ' C1 ': 3, ' O2 ': 1,
                                  ' C2 ': 4, ' C3 ': 3, ' O3 ': 2,
                                  ' O4 ': 1, ' H1 ': 1, ' H2 ': 1,
                                  ' H3 ': 1}, \
            'Unexpected pairing between atom names and degrees'

    def test_pdb_parsers(self):
        """It checks that the PDB parsers from RDKit are correctly working."""

        from rdkit.Chem.rdmolfiles import MolToPDBBlock
        from peleffy.utils.toolkits import RDKitToolkitWrapper
        from peleffy.utils import get_data_file_path

        wrapper = RDKitToolkitWrapper()

        pdb_path = get_data_file_path('ligands/benzene.pdb')
        with open(pdb_path) as f:
            pdb_block = f.read()

        rdkit_mol1 = wrapper.from_pdb(pdb_path)
        rdkit_mol2 = wrapper.from_pdb_block(pdb_block)

        block1 = MolToPDBBlock(rdkit_mol1)
        block2 = MolToPDBBlock(rdkit_mol2)

        assert block1 == block2, 'Unexpected pair of RDKit molecules'

    def test_dihedral_angle(self):
        """It checks that the dihedral angle calculator works well."""

        from peleffy.topology import Molecule
        from peleffy.utils.toolkits import RDKitToolkitWrapper
        from peleffy.utils import get_data_file_path

        wrapper = RDKitToolkitWrapper()

        pdb_path = get_data_file_path('ligands/trimethylglycine.pdb')
        m = Molecule(pdb_path)
        dihedral_degrees = wrapper.get_dihedral(m, 2, 1, 4, 5, units="degrees")
        dihedral_rad = wrapper.get_dihedral(m, 2, 1, 4, 5)
        np.testing.assert_almost_equal(dihedral_degrees, -176.348, decimal=2)
        np.testing.assert_almost_equal(dihedral_degrees, np.rad2deg(dihedral_rad), decimal=3)

    def test_dihedral_angle_2(self):
        """It checks that the dihedral angle calculator works well."""

        from peleffy.topology import Molecule
        from peleffy.utils.toolkits import RDKitToolkitWrapper
        from peleffy.utils import get_data_file_path

        wrapper = RDKitToolkitWrapper()

        pdb_path = get_data_file_path('ligands/trimethylglycine.pdb')
        m = Molecule(pdb_path)
        dihedral_degrees = wrapper.get_dihedral(m, 17, 4, 5, 6, units="degrees")
        dihedral_rad = wrapper.get_dihedral(m, 17, 4, 5, 6)
        np.testing.assert_almost_equal(dihedral_degrees, 54.828, decimal=2)
        np.testing.assert_almost_equal(dihedral_degrees, np.rad2deg(dihedral_rad), decimal=3)

    def test_rmsd(self):
        """It checks that the rmsd calculator works well."""

        from peleffy.topology import Molecule
        from peleffy.utils.toolkits import RDKitToolkitWrapper
        from peleffy.utils import get_data_file_path

        wrapper = RDKitToolkitWrapper()

        pdb_path = get_data_file_path('ligands/trimethylglycine.pdb')
        m = Molecule(pdb_path)
        pdb_path2 = get_data_file_path('ligands/trimethylglycine_moved.pdb')
        m2 = Molecule(pdb_path2)
        np.testing.assert_almost_equal(wrapper.get_rmsd(m, m2), 0.3346, decimal=3)



"""
This module handles all classes and functions related with molecular
representations.
"""


from .rotamer import MolecularGraph, MolecularGraphWithConstrainedCore
from peleffy.utils.toolkits import (RDKitToolkitWrapper,
                                    OpenForceFieldToolkitWrapper)
from peleffy.utils import Logger


class Molecule(object):
    """
    It represent wraps up all the tools to parameterize a molecule with
    the OpenForceField toolkit for PELE.
    """

    def __init__(self, path=None, smiles=None, pdb_block=None,
                 rotamer_resolution=30,exclude_terminal_rotamers=True, name='',
                 tag='UNK', connectivity_template=None, core_constraints=[],
                 allow_undefined_stereo=False, fix_pdb=True):
        """
        It initializes a Molecule object through a PDB file or a SMILES
        tag.

        Parameters
        ----------
        path : str
            The path to a PDB with the molecule structure
        smiles : str
            The smiles tag
        rotamer_resolution : float
            The resolution in degrees to discretize the rotamer's
            conformational space. Default is 30
        exclude_terminal_rotamers : bool
            Whether to exclude terminal rotamers when generating the
            rotamers library  or not
        name : str
            The molecule name
        tag : str
            The molecule tag. It must be a 3-character string
        connectivity_template : an rdkit.Chem.rdchem.Mol object
            A molecule represented with RDKit to use when assigning the
            connectivity of this Molecule object
        core_constraints : list[int or str]
            It defines the list of atoms to constrain in the core, thus,
            the core will be forced to contain them. Atoms can be specified
            through integers that match the atom index or strings that
            match with the atom PDB name
        allow_undefined_stereo : bool
            Whether to allow a molecule with undefined stereochemistry
            to be defined or try to assign the stereochemistry and
            raise a complaint if not possible. Default is False
        fix_pdb : bool
            Activates or deactivate the PDB fixer that is executed
            prior parsing it

        Examples
        --------

        Load a molecule from a PDB file and parameterize it with Open
        Force Field

        >>> from peleffy.topology import Molecule

        >>> molecule = Molecule('molecule.pdb')
        >>> molecule.parameterize('openff_unconstrained-1.2.0.offxml')

        Load a molecule using a SMILES tag and parameterize it with Open
        Force Field

        >>> from peleffy.topology import Molecule

        >>> molecule = Molecule(smiles='Cc1ccccc1')

        Load a molecule usign a PDB file (without connectivity) and assign
        the missing connectivity from an RDKit template (e.g. obtained
        from qcportal and the Open Force Field Toolkit)

        >>> import qcportal as ptl
        >>> from openforcefield.topology import Molecule as OFFMolecule

        >>> ds = client.get_collection('OptimizationDataset',
                                       'Kinase Inhibitors: WBO Distributions')
        >>> entry = ds.get_entry(ds.df.index[0])
        >>> mol_record = OFFMolecule.from_qcschema(entry)
        >>> template = mol_record.to_rdkit()

        >>> from peleffy.topology import Molecule

        >>> molecule = Molecule('PDB_without_connectivity.pdb',
                                template=template)

        Load a molecule and create its rotamer library template with
        a core constraint

        >>> from peleffy.topology import Molecule
        >>> from peleffy.topology import RotamerLibrary

        >>> molecule = Molecule(smiles='CCCC', name='butane', tag='BUT',
                                exclude_terminal_rotamers=False,
                                core_constraints=[0, ])

        >>> rotamer_library = RotamerLibrary(mol)
        >>> rotamer_library.to_file('butz')

        Deactivate OpenFF Toolkit's stereochemistry checking when
        loading a molecule

        >>> from peleffy.topology import Molecule

        >>> molecule = Molecule('molecule_with_undefined_stereochemistry.pdb',
                                allow_undefined_stereo=True)

        Display the molecular representation in a Jupyter Notebook

        >>> from peleffy.topology import Molecule

        >>> molecule = Molecule(smiles='CCCC', name='butane', tag='BUT')

        >>> display(molecule)

        """
        self._name = name
        self._tag = tag
        self._rotamer_resolution = rotamer_resolution
        self._exclude_terminal_rotamers = exclude_terminal_rotamers
        self._connectivity_template = connectivity_template
        self._core_constraints = core_constraints
        self._allow_undefined_stereo = allow_undefined_stereo
        self._fix_pdb = fix_pdb

        # Deactivate OpenForceField toolkit warnings
        import logging
        logging.getLogger().setLevel(logging.ERROR)
        if isinstance(path, str):
            from pathlib import Path
            extension = Path(path).suffix
            extension = extension.strip('.')
            if extension == 'pdb':
                self._initialize_from_pdb(path)
            else:
                raise ValueError(
                    '{} is not a valid extension'.format(extension))
        elif isinstance(smiles, str):
            self._initialize_from_smiles(smiles)
        elif isinstance(pdb_block, str):
            self._initialize_from_pdb_block(pdb_block)
        else:
            self._initialize()

        self._build_rotamers()

    def _initialize(self):
        """It initializes an empty molecule."""
        self._rdkit_molecule = None
        self._off_molecule = None
        self._rotamers = None
        self._graph = None

    def _pdb_checkup(self, path):
        """
        Safety check for PDB files in order to properly handle exceptions
        related with its format prior running the parser.

        Parameters
        ----------
        path : str
            The path to a PDB with the molecule structure
        """

        # Parse PDB file
        atom_id, res_name, res_id = ([] for i in range(3))
        connectivity = False
        with open(path) as pdb_file:
            for line in pdb_file:
                if line.startswith('ATOM') or line.startswith('HETATM'):
                    atom_id.append(line[12:16])
                    res_name.append(line[17:20])
                    res_id.append(line[22:26])
                if line.startswith('CONECT'):
                    connectivity = True

        # Handle exceptions related with the PDB file format
        if not res_id[:-1] == res_id[1:]:
            raise Exception(
                'A single ligand with immutable residue ids is expected')
        if not res_name[:-1] == res_name[1:]:
            raise Exception(
                'A single ligand with immutable residue names is expected')
        if not len(atom_id) == len(set(atom_id)):
            raise Exception(
                'Ligand in input PDB has no unique atom names')
        if not connectivity and self.connectivity_template is None:
            log = Logger()
            log.warning(
                "Warning: input PDB has no information about the "
                + "connectivity and this could result in an unexpected "
                + "bond assignment")

    def _read_and_fix_pdb(self, path):
        """
        It reads the input PDB file returns the corresponding PDB block.
        It also applies some modifications, in case it requires some
        fixing prior running the parser.

        Parameters
        ----------
        path : str
            The path to a PDB with the molecule structure

        Returns
        -------
        pdb_block : str
            The corresponding PDB block, with applied fixes if required
        """
        log = Logger()

        # Skip PDB fixing if it has been deactivated
        if not self.fix_pdb:
            with open(path) as pdb_file:
                pdb_block = pdb_file.read()

            return pdb_block

        # Fix PDB
        missing_element = False
        any_fail = False
        pdb_block = ''
        with open(path) as pdb_file:
            for line in pdb_file:
                if line.startswith('ATOM') or line.startswith('HETATM'):
                    if len(line) < 78 or line[76:78] == '  ':
                        missing_element = True
                        atom_name = line[12:16]
                        # Try to infer element from atom name
                        inferred_element = ''.join([c for c in atom_name
                                                    if not c.isdigit()
                                                    and c != ' '])

                        # Format properly the element identifier
                        if len(inferred_element) == 1:
                            inferred_element = inferred_element.upper()
                        elif len(inferred_element) == 2:
                            inferred_element = inferred_element[0].upper() + \
                                inferred_element[1].lower()
                        else:
                            # We were expecting an element identifier of 1 or 2 chars
                            any_fail = True
                            break

                        # Remove line breaks, if any
                        line = line.strip()

                        # Fill a short line with white spaces
                        while(len(line) < 79):
                            line += ' '

                        # Add element to line (right-justified)
                        line = line[:76] + '{:>2s}'.format(inferred_element) \
                            + line[79:] + '\n'

                pdb_block += line

        if missing_element:
            log.warning(
                "Warning: input PDB has no information about atom "
                + "elements and they were inferred from atom names. "
                + "Please, verify that the resulting elements are "
                + "correct")

        if any_fail:
            log.error("Error: PDB could not be fixed")
            with open(path) as pdb_file:
                pdb_block = pdb_file.read()

        return pdb_block

    def _initialize_from_pdb(self, path):
        """
        It initializes a molecule with the molecule structure read from
        a PDB file.

        Parameters
        ----------
        path : str
            The path to a PDB with the molecule structure
        """
        logger = Logger()
        logger.info(' - Initializing molecule from PDB')
        self._initialize()

        # Validate PDB
        self._pdb_checkup(path)

        # Read and fix PDB
        pdb_block = self._read_and_fix_pdb(path)

        logger.info('   - Loading molecule from RDKit')
        rdkit_toolkit = RDKitToolkitWrapper()
        self._rdkit_molecule = rdkit_toolkit.from_pdb_block(pdb_block)

        # Use RDKit template, if any, to assign the connectivity to
        # the current Molecule object
        if self.connectivity_template is not None:
            logger.info('   - Assigning connectivity from template')
            rdkit_toolkit.assign_connectivity_from_template(self)

        if not self.allow_undefined_stereo:
            # RDKit must generate stereochemistry specifically from 3D coords
            logger.info('   - Assigning stereochemistry from 3D coordinates')
            rdkit_toolkit.assign_stereochemistry_from_3D(self)

        # Set molecule name according to PDB name
        if self.name == '':
            from pathlib import Path
            name = Path(path).stem
            logger.info('   - Setting molecule name to \'{}\''.format(name))
            self.set_name(name)

        # Set molecule tag according to PDB's residue name
        if self.tag == 'UNK':
            tag = rdkit_toolkit.get_residue_name(self)
            logger.info('   - Setting molecule tag to \'{}\''.format(tag))
            self.set_tag(tag)

        logger.info('   - Representing molecule with the Open Force Field '
                    + 'Toolkit')
        openforcefield_toolkit = OpenForceFieldToolkitWrapper()
        self._off_molecule = openforcefield_toolkit.from_rdkit(self)

    def _initialize_from_smiles(self, smiles):
        """
        It initializes a molecule from a SMILES tag.

        Parameters
        ----------
        smiles : str
            The SMILES tag to construct the molecule structure with
        """
        logger = Logger()
        logger.info(' - Initializing molecule from a SMILES tag')
        self._initialize()

        logger.info('   - Loading molecule from RDKit')
        rdkit_toolkit = RDKitToolkitWrapper()
        self._rdkit_molecule = rdkit_toolkit.from_smiles(smiles)

        # TODO not sure if stereochemistry assignment from 3D is still necessary
        # RDKit must generate stereochemistry specifically from 3D coords
        # rdkit_toolkit.assign_stereochemistry_from_3D(self)

        # Set molecule name according to the SMILES tag
        if self.name == '':
            logger.info('   - Setting molecule name to \'{}\''.format(smiles))
            self.set_name(smiles)

        logger.info('   - Representing molecule with the Open Force Field '
                    + 'Toolkit')
        openforcefield_toolkit = OpenForceFieldToolkitWrapper()
        self._off_molecule = openforcefield_toolkit.from_rdkit(self)

    def _initialize_from_pdb_block(self, pdb_block):
        """
        It initializes a molecule with the molecule structure fetch in a PDB
        block.

        Parameters
        ----------
        pdb_block : str
            PDB block with the molecule structure
        """
        logger = Logger()
        logger.info(' - Initializing molecule from PDB')
        self._initialize()

        logger.info('   - Loading molecule from RDKit')
        rdkit_toolkit = RDKitToolkitWrapper()
        self._rdkit_molecule = rdkit_toolkit.from_pdb_block(pdb_block)

        # Use RDKit template, if any, to assign the connectivity to
        # the current Molecule object
        if self.connectivity_template is not None:
            logger.info('   - Assigning connectivity from template')
            rdkit_toolkit.assign_connectivity_from_template(self)

        if not self.allow_undefined_stereo:
            # RDKit must generate stereochemistry specifically from 3D coords
            logger.info('   - Assigning stereochemistry from 3D coordinates')
            rdkit_toolkit.assign_stereochemistry_from_3D(self)

        # Set molecule tag according to PDB's residue name
        if self.tag == 'UNK':
            tag = rdkit_toolkit.get_residue_name(self)
            logger.info('   - Setting molecule tag to \'{}\''.format(tag))
            self.set_tag(tag)

        logger.info('   - Representing molecule with the Open Force Field '
                    + 'Toolkit')
        openforcefield_toolkit = OpenForceFieldToolkitWrapper()
        self._off_molecule = openforcefield_toolkit.from_rdkit(self)

    def _build_rotamers(self):
        """It builds the rotamers of the molecule."""
        logger = Logger()
        if self.off_molecule and self.rdkit_molecule:
            logger.info(' - Generating rotamer library')

            if len(self.core_constraints) != 0:
                self._graph = MolecularGraphWithConstrainedCore(
                    self, self.core_constraints)
                if len(self.core_constraints) == 1:
                    logger.info('   - Core forced to contain atom: '
                                + self._graph.constraint_names[0])
                else:
                    logger.info('   - Core forced to contain atoms: '
                                + ', '.join(atom_name.strip() for atom_name
                                            in self._graph.constraint_names))
            else:
                self._graph = MolecularGraph(self)
                logger.info('   - Core set to the center of the molecule')

            self._rotamers = self._graph.get_rotamers()

    def set_name(self, name):
        """
        It sets the name of the molecule.

        Parameters
        ----------
        name : str
            The name to set to the molecule
        """
        assert isinstance(name, str), 'Invalid type for a name, it must be ' \
            + 'a string'

        self._name = name

    def set_tag(self, tag):
        """
        It sets the tag of the molecule. It must be a 3-character string.

        Parameters
        ----------
        tag : str
            The tag to set to the molecule. It must be a 3-character string
        """
        # Some previous checks
        assert len(tag) == 3, 'Invalid tag length, it must be a ' \
            + '3-character string'
        assert isinstance(tag, str), 'Invalid type for a tag, it must be ' \
            + 'a string'

        self._tag = tag.upper()

    def get_pdb_atom_names(self):
        """
        It returns the PDB atom names of all the atoms in the molecule.

        Returns
        -------
        pdb_atom_names : list[str]
            The PDB atom names of all the atoms in this Molecule object
        """
        rdkit_toolkit = RDKitToolkitWrapper()

        return rdkit_toolkit.get_atom_names(self)

    def to_pdb_file(self, path):
        """
        It writes the molecule to a PDB file.

        Parameters
        ----------
        path : str
            Path to write to
        """
        rdkit_toolkit = RDKitToolkitWrapper()
        rdkit_toolkit.to_pdb_file(self, path)

    def set_conformer(self, conformer):
        """
        It sets a new conformer to the molecule.

        Parameters
        ----------
        conformer : an RDKit.Chem.rdchem.Conformer object
            The conformer to set to the molecule
        """
        rdkit_toolkit = RDKitToolkitWrapper()
        rdkit_toolkit.set_conformer(self, conformer)

    def get_conformer(self):
        """
        It gets the current conformer of the molecule.

        Returns
        -------
        conformer : numpy.ndarray
            The array of 3D coordinates of all the atoms in the molecule
        """
        rdkit_toolkit = RDKitToolkitWrapper()

        return rdkit_toolkit.get_coordinates(self)

    @staticmethod
    def from_rdkit(rdkit_molecule, rotamer_resolution=30,
                   exclude_terminal_rotamers=True, name='', tag='UNK',
                   connectivity_template=None, core_constraints=[],
                   allow_undefined_stereo=False):
        """
        It initializes and returns a peleffy Molecule representation
        from an RDKit molecular representation.

        Parameters
        ----------
        rdkit_molecule : an rdkit.Chem.rdchem.Mol object
            The RDKit's Molecule object to use to initialize a peleffy
            Molecule object
        rotamer_resolution : float
            The resolution in degrees to discretize the rotamer's
            conformational space. Default is 30
        exclude_terminal_rotamers : bool
            Whether to exclude terminal rotamers when generating the
            rotamers library  or not
        name : str
            The molecule name
        tag : str
            The molecule tag. It must be a 3-character string
        connectivity_template : an rdkit.Chem.rdchem.Mol object
            A molecule represented with RDKit to use when assigning the
            connectivity of this Molecule object
        core_constraints : list[int or str]
            It defines the list of atoms to constrain in the core, thus,
            the core will be forced to contain them. Atoms can be specified
            through integers that match the atom index or strings that
            match with the atom PDB name
        allow_undefined_stereo : bool
            Whether to allow a molecule with undefined stereochemistry
            to be defined or try to assign the stereochemistry and
            raise a complaint if not possible. Default is False

        Returns
        -------
        molecule : an peleffy.topology.Molecule
            The resulting peleffy's Molecule object

        Examples
        --------

        Load a molecule from an RDKit molecular representation

        >>> from rdkit import Chem

        >>> rdkit_molecule = Chem.MolFromPDBFile(pdb_path, removeHs=False)

        >>> from peleffy.topology import Molecule

        >>> molecule = Molecule.from_rdkit(rdkit_molecule)

        """
        molecule = Molecule(
            rotamer_resolution=30,
            exclude_terminal_rotamers=exclude_terminal_rotamers,
            name=name, tag=tag,
            connectivity_template=connectivity_template,
            core_constraints=core_constraints,
            allow_undefined_stereo=allow_undefined_stereo)

        logger = Logger()

        logger.info(' - Initializing molecule from an RDKit '
                    + 'molecular representation')
        molecule._initialize()
        molecule._rdkit_molecule = rdkit_molecule

        logger.info('   - Representing molecule with the Open Force Field '
                    + 'Toolkit')
        openforcefield_toolkit = OpenForceFieldToolkitWrapper()
        molecule._off_molecule = openforcefield_toolkit.from_rdkit(molecule)

        molecule._build_rotamers()

        return molecule

    @staticmethod
    def from_openff(openff_molecule, rotamer_resolution=30,
                    exclude_terminal_rotamers=True, name='', tag='UNK',
                    connectivity_template=None, core_constraints=[],
                    allow_undefined_stereo=False):
        """
        It initializes and returns a peleffy Molecule representation
        from an OpenForceField molecular representation.

        Parameters
        ----------
        openff_molecule : an openforcefield.topology.Molecule object
            The OpenForceField's Molecule to use to initialize a peleffy
            Molecule object
        rotamer_resolution : float
            The resolution in degrees to discretize the rotamer's
            conformational space. Default is 30
        exclude_terminal_rotamers : bool
            Whether to exclude terminal rotamers when generating the
            rotamers library  or not
        name : str
            The molecule name
        tag : str
            The molecule tag. It must be a 3-character string
        connectivity_template : an rdkit.Chem.rdchem.Mol object
            A molecule represented with RDKit to use when assigning the
            connectivity of this Molecule object
        core_constraints : list[int or str]
            It defines the list of atoms to constrain in the core, thus,
            the core will be forced to contain them. Atoms can be specified
            through integers that match the atom index or strings that
            match with the atom PDB name
        allow_undefined_stereo : bool
            Whether to allow a molecule with undefined stereochemistry
            to be defined or try to assign the stereochemistry and
            raise a complaint if not possible. Default is False

        Returns
        -------
        molecule : an peleffy.topology.Molecule
            The resulting peleffy's Molecule object

        Examples
        --------

        Load a molecule from an RDKit molecular representation

        >>> from rdkit import Chem

        >>> rdkit_molecule = Chem.MolFromPDBFile(pdb_path)

        >>> from peleffy.topology import Molecule

        >>> molecule = Molecule.from_rdkit(rdkit_molecule)

        """
        if name == '':
            name = openff_molecule.name

        molecule = Molecule(
            rotamer_resolution=30,
            exclude_terminal_rotamers=exclude_terminal_rotamers,
            name=name, tag=tag,
            connectivity_template=connectivity_template,
            core_constraints=core_constraints,
            allow_undefined_stereo=allow_undefined_stereo)

        logger = Logger()

        logger.info(' - Initializing molecule from an OpenFF '
                    + 'molecular representation')
        molecule._initialize()
        molecule._off_molecule = openff_molecule

        logger.info('   - Generating RDKit molecular representation with '
                    + 'the Open Force Field Toolkit')
        molecule._rdkit_molecule = openff_molecule.to_rdkit()

        molecule._build_rotamers()

        return molecule

    @property
    def rotamer_resolution(self):
        """
        The resolution to be used when generating the rotamers library.

        Returns
        -------
        rotamer_resolution : float
            The resolution in degrees to discretize the rotamer's
            conformational space
        """
        return self._rotamer_resolution

    @property
    def exclude_terminal_rotamers(self):
        """
        The behavior when handling terminal rotamers when generating the
        rotamers library.

        Returns
        -------
        exclude_terminal_rotamers : bool
            Whether to exclude terminal rotamers when generating the
            rotamers library  or not
        """
        return self._exclude_terminal_rotamers

    @property
    def connectivity_template(self):
        """
        The template containing the correct connectivity for this Molecule
        object.

        Returns
        -------
        connectivity_template : an rdkit.Chem.rdchem.Mol object
            A molecule represented with RDKit to use when assigning the
            connectivity of this Molecule object
        """
        return self._connectivity_template

    @property
    def allow_undefined_stereo(self):
        """
        Whether to allow a molecule with undefined stereochemistry
        to be defined or try to assign the stereochemistry and
        raise a complaint if not possible

        Returns
        -------
        allow_undefined_stereo : bool
            The current configuration towards the stereochemistry
            behaviour of this molecule
        """
        return self._allow_undefined_stereo

    @property
    def fix_pdb(self):
        """
        It activates or deactivates the PDB fixer prior parsing the PDB
        file.

        Returns
        -------
        fix_pdb : bool
            The current activation state of the PDB fixer
        """
        return self._fix_pdb

    @property
    def core_constraints(self):
        """
        The list of indices or PDB names of the atoms to constraint to
        the core when building the rotamers.

        Returns
        -------
        core_constraints : list[int or str]
            The list of indices or PDB names of the atoms to constrain
        """
        return self._core_constraints

    @property
    def off_molecule(self):
        """
        The OpenForceField's molecule instance linked to the molecule.

        Returns
        -------
        off_molecule : an openforcefield.topology.Molecule
            The OpenForceField's molecule linked to this Molecule object
        """
        return self._off_molecule

    @property
    def rdkit_molecule(self):
        """
        The RDKit's molecule instance linked to the molecule.

        Returns
        -------
        rdkit_molecule : an rdkit.Chem.rdchem.Mol object
            The RDKit's molecule linked to this Molecule object
        """
        return self._rdkit_molecule

    @property
    def rotamers(self):
        """
        The list of rotamers of the molecule.

        Returns
        -------
        rotamers : list[list]
            The list of rotamers grouped by the branch they belong to
        """
        return self._rotamers

    @property
    def name(self):
        """
        Molecule's name.

        Returns
        -------
        name : str
            The name of this Molecule object
        """
        return self._name

    @property
    def tag(self):
        """
        Molecule's tag.

        Returns
        -------
        tag : str
            The tag of this Molecule object
        """
        return self._tag

    @property
    def graph(self):
        """
        The topological graph of the molecule.

        Returns
        -------
        graph : an peleffy.topology.rotamer.MolecularGraph object
            The topological graph of this Molecule object.
        """
        return self._graph

    def _ipython_display_(self):
        """
        It returns a RDKit molecule with an embeded 2D representation.

        Returns
        -------
        representation_2D : a IPython display object
            It is displayable RDKit molecule with an embeded 2D
            representation
        """
        from IPython.display import display

        # Get 2D molecular representation
        rdkit_toolkit = RDKitToolkitWrapper()
        representation = rdkit_toolkit.get_2D_representation(self)

        # Get its image
        image = rdkit_toolkit.draw_molecule(representation)

        return display(image)



"""
This module contains a variety of helpful tools for tests.
"""


import numpy as np
from simtk import unit

from peleffy.forcefield.forcefield import _BaseForceField


SET_OF_LIGAND_PATHS = ['ligands/benzene.pdb', 'ligands/toluene.pdb',
                       'ligands/oleic_acid.pdb', 'ligands/propionic_acid.pdb',
                       'ligands/trimethylglycine.pdb', 'ligands/ethylene.pdb']


def apply_PELE_dihedral_equation(proper, x):
    """
    Given an x, it applies the PELE's dihedral equation to obtain a y.

    Parameters
    ----------
    proper : a peleffy.topology.Proper
        The proper whose parameters will be applied to equation
    x : float
        Equation's x value
    """
    return proper.constant * (1 + proper.prefactor
                              * np.cos(proper.periodicity * x))


def apply_OFF_dihedral_equation(proper, x):
    """
    Given an x, it applies the Open Force Field's dihedral equation to obtain
    a y.

    Parameters
    ----------
    proper : a peleffy.topology.Proper
        The proper whose parameters will be applied to equation
    x : float
        Equation's x value
    """
    return proper.k * (1 + np.cos(proper.periodicity * x - proper.phase))


def check_CHO_charges(parameters):
    """
    It checks whether C, H and O atoms in a molecule have reasonable
    partial charges assigned.

    Parameters
    ----------
    parameters : a peleffy.forcefield.parameters.BaseParameterWrapper object
        The parameter wrapper containing the generated parameters
        along with the partial charges to check

    Raises
    ------
    AssertionError in case that the checking fails
    ValueError if an unexpected element is found in the molecule
    """

    for name, charge in zip(parameters['atom_names'], parameters['charges']):
        charge = charge.value_in_unit(unit.elementary_charge)

        if 'C' in name:
            assert charge < 1.0 and charge > -0.23, \
                'Presumably not a reasonable charge for a C: ' \
                + '{}'.format(charge)
        elif 'H' in name:
            assert charge > 0 and charge < 0.2, \
                'Presumable not a reasonable charge for a H: ' \
                + '{}'.format(charge)
        elif 'O' in name:
            assert charge < -0.5 and charge > -1.0, \
                'Presumable not a reasonable charge for a O: ' \
                + '{}'.format(charge)
        else:
            raise ValueError('Unknown atom name')


def compare_dicts(dict1, dict2):
    """
    Given two dictionaries, it compares them and complains about any
    disagreement.

    Parameters
    ----------
    dict1 : str
        First dictionary to compare
    dict2 : str
        Second dictionary to compare

    Raises
    ------
        AssertionError
            If any difference is found between dictionaries
    """

    assert len(dict1) == len(dict2), 'Number of keys does not match, ' \
        + 'dictionary1: {}, dictionary2: {}'.format(len(dict1), len(dict2))

    for key in dict1.keys():
        assert key in dict2.keys(), 'Key \'{}\' from '.format(key) \
            + 'directory1 not found in dictionary2'

    for key in dict2.keys():
        assert key in dict1.keys(), 'Key \'{}\' from '.format(key) \
            + 'dictionary2 not found in dictionary1'

    for key, value in dict1.items():
        assert value == dict2[key], 'Value for key \'{}\' '.format(key) \
            + 'does not match between dictionaries, ' \
            + 'dictionary1: {}, dictionary2: {}'.format(value, dict2[key])


def merge_dicts(*dict_args):
    """
    Given any number of dictionaries, shallow copy and merge into a new dict,
    precedence goes to key-value pairs in latter dictionaries.

    Parameters
    ----------
    **dict_args : dict
        Dictionary to merge

    Returns
    -------
    merged_dict : str
        Merged dictionary
    """
    merged_dict = {}
    for dictionary in dict_args:
        merged_dict.update(dictionary)
    return merged_dict


def check_parameters(topology, expected_nonbonding=None,
                     expected_bonds=None, expected_angles=None,
                     expected_propers=None, expected_impropers=None):
    """
    It checks the current parameters of the molecule.

    Parameters
    ----------
    topology : a Topology object
        The molecular topology representation to check
    expected_nonbonding : list[list]
        The list of expected nonbonding parameters
    expected_bonds : list[list]
        The list of expected bond parameters
    expected_angles : list[list]
        The list of expected angle parameters
    expected_propers : list[list]
        The list of expected proper parameters
    expected_impropers : list[list]
        The list of expected improper parameters
    """

    from peleffy.template.impact import (WritableAtom,
                                         WritableBond,
                                         WritableAngle,
                                         WritableProper,
                                         WritableImproper)

    if expected_nonbonding is not None:
        assert len(topology.atoms) == len(expected_nonbonding), \
            'Invalid number of nonbonding terms'

        for atom in topology.atoms:
            w_atom = WritableAtom(atom)
            w_parameters = [w_atom.index, w_atom.parent.index, w_atom.core,
                            w_atom.OPLS_type, w_atom.PDB_name,
                            w_atom.unknown, w_atom.sigma, w_atom.epsilon,
                            w_atom.charge, w_atom.born_radius,
                            w_atom.SASA_radius, w_atom.nonpolar_gamma,
                            w_atom.nonpolar_alpha]
            assert w_parameters in expected_nonbonding, \
                'Invalid writable nonbonding parameters ' \
                + '{}'.format(w_parameters)

    if expected_bonds is not None:
        assert len(topology.bonds) == len(expected_bonds), \
            'Invalid number of bond terms'

        for bond in topology.bonds:
            w_bond = WritableBond(bond)
            w_parameters = [attr[1] for attr in list(w_bond)]
            assert w_parameters in expected_bonds, \
                'Invalid writable bond parameters {}'.format(w_parameters)

    if expected_angles is not None:
        assert len(topology.angles) == len(expected_angles), \
            'Invalid number of angle terms'

        for angle in topology.angles:
            w_angle = WritableAngle(angle)
            w_parameters = [attr[1] for attr in list(w_angle)]
            assert w_parameters in expected_angles, \
                'Invalid writable angle parameters {}'.format(w_parameters)

    if expected_propers is not None:
        assert len(topology.propers) == len(expected_propers), \
            'Invalid number of proper terms'

        for proper in topology.propers:
            w_proper = WritableProper(proper)
            w_parameters = [attr[1] for attr in list(w_proper)]
            assert w_parameters in expected_propers, \
                'Invalid writable proper parameters {}'.format(w_parameters)

    if expected_impropers is not None:
        assert len(topology.impropers) == len(expected_impropers), \
            'Invalid number of improper terms'

        for improper in topology.impropers:
            w_improper = WritableImproper(improper)
            w_parameters = [attr[1] for attr in list(w_improper)]
            assert w_parameters in expected_impropers, \
                'Invalid writable improper parameters {}'.format(w_parameters)


def compare_files(file1, file2):
    """
    It compares two files line by line and gives an AssertionError if there
    is any difference.

    Parameters
    ----------
    file1 : str
        Path to the first file to compare
    file2 : str
        Path to the second file to compare

    Raises
    ------
        AssertionError
            If any difference is found between files
    """
    with open(file1, 'r') as f1:
        lines1 = [line for line in f1.readlines()
                  if not line.startswith("*")]
    with open(file2, 'r') as f2:
        lines2 = [line for line in f2.readlines()
                  if not line.startswith("*")]
    assert len(lines1) == len(lines2), \
        'Number of lines do not match: ' \
        + str(len(lines1)) + ' and ' + str(len(lines2))
    for i, (line1, line2) in enumerate(zip(lines1, lines2)):
        assert line1 == line2, \
            'Found different lines at line {}:'.format(i) + '\n' + line1 + line2


def compare_files_without_order(file1, file2):
    """
    It compares two files line by line and gives an AssertionError if there
    is any difference. The order of the lines is not taken into account.

    Parameters
    ----------
    file1 : str
        Path to the first file to compare
    file2 : str
        Path to the second file to compare

    Raises
    ------
        AssertionError
            If any difference is found between files
    """
    with open(file1, 'r') as f1:
        lines1 = [line for line in f1.readlines()
                  if not line.startswith("*")]
    with open(file2, 'r') as f2:
        lines2 = [line for line in f2.readlines()
                  if not line.startswith("*")]

    assert len(lines1) == len(lines2), \
        'Number of lines do not match: ' \
        + str(len(lines1)) + ' and ' + str(len(lines2))

    for i, (line1, line2) in enumerate(zip(lines1, lines2)):
        assert line1 in lines2, \
            'Line not found in reference template {}:' \
            .format(i) + '\n' + line1


def compare_blocks(block1, block2, line_indices):
    """
    It compares two PDB blocks and gives an AssertionError if there is any
    difference.

    Parameters
    ----------
    block1 : str
        First block to compare
    block2 : str
        Second block to compare
    line_indices : tuple[int, int]
        The line indices to compare

    Raises
    ------
        AssertionError
            If any difference is found between blocks
    """

    idx1, idx2 = line_indices

    lines1 = list()
    lines2 = list()

    for line in block1.split('\n'):
        if line.startswith('REMARK'):
            continue
        if line.startswith('TITLE'):
            continue
        if line.startswith('CONECT'):
            continue
        if line.startswith('CRYST'):
            continue
        if line.startswith('END'):
            continue
        if line.startswith('MODEL'):
            continue
        if line.startswith('ENDMDL'):
            continue
        if line.startswith('TER'):
            continue
        if line == '':
            continue
        lines1.append(line)

    for line in block2.split('\n'):
        if line.startswith('REMARK'):
            continue
        if line.startswith('TITLE'):
            continue
        if line.startswith('CONECT'):
            continue
        if line.startswith('CRYST'):
            continue
        if line.startswith('END'):
            continue
        if line.startswith('MODEL'):
            continue
        if line.startswith('ENDMDL'):
            continue
        if line.startswith('TER'):
            continue
        if line == '':
            continue
        lines2.append(line)

    assert len(lines1) == len(lines2), \
        'Number of lines do not match: ' \
        + str(len(lines1)) + ' and ' + str(len(lines2))

    for line1, line2 in zip(lines1, lines2):
        print(line1[idx1:idx2] + 'X')
        print(line2[idx1:idx2] + 'X')

        assert line1[idx1:idx2] == line2[idx1:idx2], \
            'Found different lines:\n' + line1 + line2


def parameterize_opls2005(opls2005, molecule, ffld_file,
                          charge_method='OPLS2005'):
    """
    This is a workaround to parameterize an OPLS2005 force field
    using a precomputed ffld_file. Thus, we do not require the
    Schrodinger dependency.

    Parameters
    ----------
    opls2005 : an OPLS2005ForceField object
        The OPLS2005 force field to parameterize with the ffld_file
    molecule : a peleffy.topology.Molecule
        The peleffy's Molecule object to parameterize with the ffld file
    ffld_file : str
        The path to the precomputed ffld file from where the parameters
        will be extracted
    charge_method : str
        The charge method to employ to calculate partial charges.
        Default is 'OPLS2005'

    Returns
    -------
    parameters : a peleffy.forcefield.parameters.BaseParameterWrapper object
        The parameter wrapper containing the parameters generated
        with the precomputed ffld_file
    """

    from peleffy.forcefield.parameters import OPLS2005ParameterWrapper

    with open(ffld_file) as f:
        ffld_output = f.read()

    parameters = OPLS2005ParameterWrapper.from_ffld_output(molecule,
                                                           ffld_output)
    # Assign partial charges using the charge calculator object
    charge_calculator = opls2005._get_charge_calculator(charge_method,
                                                        molecule)
    charge_calculator.assign_partial_charges(parameters)

    return parameters


def parameterize_openffopls2005(openffopls2005, molecule, ffld_file):
    """
    This is a workaround to parameterize an OpenFFOPLS2005 force field
    using a precomputed ffld_file. Thus, we do not require the
    Schrodinger dependency.

    Parameters
    ----------
    openffopls2005 : an OpenFFOPLS2005ForceField object
        The OPLS2005 force field to parameterize with the ffld_file
    molecule : a peleffy.topology.Molecule
        The peleffy's Molecule object to parameterize with the ffld file
    ffld_file : str
        The path to the precomputed ffld file from where the parameters
        will be extracted

    Returns
    -------
    parameters : a peleffy.forcefield.parameters.BaseParameterWrapper object
        The parameter wrapper containing the parameters generated
        with the precomputed ffld_file
    """

    parameters = parameterize_opls2005(openffopls2005._oplsff,
                                       molecule,
                                       ffld_file)

    # Initialize mock class
    oplsff = MockOPLS2005ForceField()
    oplsff.set_preloaded_parameters(parameters)

    # Set mock class to the OpenFFOPLS2005ForceField class
    openffopls2005._oplsff = oplsff

    print(parameters['sigmas'])

    return openffopls2005.parameterize(molecule)


class MockBaseForceField(_BaseForceField):
    """
    It is a mock class of _BaseForceField to skip Schrodinger
    dependency.
    """

    _preloaded_parameters = None

    def parameterize(self, molecule):
        """
        It parameterizes the supplied molecule.

        Parameters
        ----------
        molecule : a peleffy.topology.Molecule
            The peleffy's Molecule object to parameterize

        Returns
        -------
        parameters : a peleffy.forcefield.parameters.BaseParameterWrapper object
            The parameter wrapper containing the parameters generated
            with the current force field
        """

        # Assign parameters
        parameters = self._preloaded_parameters

        return parameters

    def set_preloaded_parameters(self, parameters):
        """
        It loads parameters to the mock class of BaseForceField.

        Parameters
        ----------
        parameters : a peleffy.forcefield.parameters.BaseParameterWrapper object
            The parameter wrapper containing the preloaded parameters
            for this mock class
        """

        self._preloaded_parameters = parameters


class MockOPLS2005ForceField(MockBaseForceField):
    """
    It is a mock class of OPLS2005ForceField to skip Schrodinger
    dependency.
    """

    _preloaded_parameters = None

    def __init__(self):
        """It initializes the OPLS2005 force field class."""
        super().__init__(self._type)

    def _get_parameters(self, molecule):
        """
        Instead of computing the parameters and returning
        """
        return self._preloaded_parameters

    def set_preloaded_parameters(self, parameters):
        """
        It loads parameters to the mock class of OPLS2005ForceField.

        Parameters
        ----------
        parameters : a peleffy.forcefield.parameters.BaseParameterWrapper object
            The parameter wrapper containing the preloaded parameters
            for this mock class
        """

        self._preloaded_parameters = parameters



"""
This module contains the tests to check peleffy's solvent.
"""

import tempfile

from peleffy.utils import get_data_file_path, temporary_cd
from peleffy.topology import Molecule, Topology
from peleffy.forcefield import OPLS2005ForceField, OpenForceField
from peleffy.solvent import OPLSOBC, OBC2


class TestSolvent(object):
    """
    It contains all the tests that validate the solvent-template generator.
    """

    def test_single_topology(self):
        """
        It tests the class that generates a OpenFFCompatibleSolvent object for
        a single topology.
        """
        from .utils import compare_dicts
        import json

        TEMPLATE_PARAMS_MAL = get_data_file_path('tests/ligandParams_MAL.txt')

        # Loads the  molecule
        molecule = Molecule(path=get_data_file_path('ligands/malonate.pdb'),
                            tag='MAL')

        # Sets forcefield and parameterizes it
        ff = OpenForceField('openff_unconstrained-1.2.1.offxml')
        parameters = ff.parameterize(molecule, charge_method='gasteiger')

        # Initializes topology
        topology = Topology(molecule, parameters)

        # Initializes solvent and gets parameters file
        solvent = OBC2(topology)
        solvent_dict = solvent.to_dict()

        # Loads reference dict from template
        with open(TEMPLATE_PARAMS_MAL, 'r') as f:
            reference_dict = json.load(f)

        # Compare the output parameters dict with the reference parameters
        compare_dicts(reference_dict, solvent_dict)

    def test_multiple_topologies(self):
        """
        It tests the class that generates a OpenFFCompatibleSolvent object for
        multiple topologies.
        """
        from .utils import compare_dicts, merge_dicts

        # Path to multiple non standard residues
        pdb_path_MAL = get_data_file_path('ligands/malonate.pdb')
        pdb_path_MET = get_data_file_path('ligands/methane.pdb')

        # Force Field to parameterize the molecules
        ff = OpenForceField('openff_unconstrained-1.2.1.offxml')

        # Topology of malonate
        mol_MAL = Molecule(path=pdb_path_MAL, tag='MAL')
        parameters_MAL = ff.parameterize(mol_MAL, charge_method='gasteiger')
        topology_MAL = Topology(mol_MAL, parameters_MAL)

        # Topology of methane
        mol_MET = Molecule(path=pdb_path_MET, tag='MET')
        parameters_MET = ff.parameterize(mol_MET, charge_method='gasteiger')
        topology_MET = Topology(mol_MET, parameters_MET)

        # List containing both topologies
        topologies = [topology_MAL, topology_MET]

        # Generate the Solvent parameters dictionaries
        solvent_MAL_dict = OBC2(topology_MAL).to_dict()
        solvent_MET_dict = OBC2(topology_MET).to_dict()
        solvent_dict = OBC2(topologies).to_dict()

        # Check that merging both single topology dicitionaries we obtain the
        # same dictionary that using multiple topologies
        compare_dicts(merge_dicts(solvent_MAL_dict['SolventParameters'],
                                  solvent_MET_dict['SolventParameters']),
                      solvent_dict['SolventParameters'])

    def test_multiple_topologies_writer(self):
        """
        It tests the class that generates a OpenFFCompatibleSolvent object for multiple topologies. It compares the outcome of the Solvent writer with
        a reference file.
        """
        from .utils import compare_dicts, parameterize_opls2005
        import json

        TEMPLATE_PARAMS = get_data_file_path('tests/ligandParams.txt')

        with tempfile.TemporaryDirectory() as tmpdir:
            with temporary_cd(tmpdir):
                path_OXO = get_data_file_path('tests/MRO_oleic/OXO.pdb')
                path_OLC = get_data_file_path('tests/MRO_oleic/OLC.pdb')

                ff = OpenForceField('openff_unconstrained-1.2.1.offxml')
                opls2005 = OPLS2005ForceField()

                # Group OXO
                m_OXO = Molecule(path_OXO)
                ffld_file = get_data_file_path('tests/OXO_ffld_output.txt')
                parameters_OXO = parameterize_opls2005(opls2005, m_OXO,
                                                       ffld_file)
                topology_OXO = Topology(m_OXO, parameters_OXO)

                # Acid oleic
                m_OLC = Molecule(path_OLC)
                parameters_OLC = ff.parameterize(m_OLC,
                                                 charge_method='gasteiger')
                topology_OLC = Topology(m_OLC, parameters_OLC)

                # Multiple topologies
                topologies = [topology_OXO, topology_OLC]
                solvent = OBC2(topologies)
                solvent.to_file('OBC_parameters.txt')

                # Loads reference dict from template
                with open(TEMPLATE_PARAMS, 'r') as f:
                    reference_dict = json.load(f)

                # Loads the generated template into a dict
                with open('OBC_parameters.txt', 'r') as f:
                    solvent_dict = json.load(f)

                # Compare the output parameters dict with the reference parameters
                compare_dicts(reference_dict, solvent_dict)

    def test_OBCOPLS_writer(self):
        """
        It test the function that writes a OPLS2005CompatibleSolvent object to
        a file compatible with PELE.
        """
        from .utils import parameterize_opls2005, compare_files_without_order

        TEMPLATE_PARAMS_ETL = get_data_file_path(
            'tests/ETL_solventParamsHCTOBC.txt')
        TEMPLATE_PARAMS_MAL = get_data_file_path(
            'tests/MAL_solventParamsHCTOBC.txt')
        TEMPLATE_PARAMS_MET = get_data_file_path(
            'tests/MET_solventParamsHCTOBC.txt')

        def test_OBCOPLS_writer_ligand(pdbfile, tag_name, ffld_name,
                                       reference_file):
            """
            Given a ligand, it tests that the output parameters file
            corresponds to the refenrece file.

            Parameters
            ----------
            pdbfile : str
                The path to the PDB of the ligand to test
            ffld_name : str
                The path to the ffld_server's output file
            reference_file : str
                The path to reference TXT file compatible with PELE
            """
            with tempfile.TemporaryDirectory() as tmpdir:
                with temporary_cd(tmpdir):

                    # Loads the  molecule
                    molecule = Molecule(get_data_file_path(pdbfile),
                                        tag=tag_name)

                    # Sets forcefield and parameterizes it
                    opls2005 = OPLS2005ForceField()
                    ffld_file = get_data_file_path(ffld_name)
                    parameters = parameterize_opls2005(opls2005,
                                                       molecule,
                                                       ffld_file)

                    # Initializes topology
                    topology = Topology(molecule, parameters)

                    # Initializes solvent and gets parameters file
                    solvent = OPLSOBC(topology)
                    solvent.to_file('OBC_parameters.txt')

                    # Compare the output file with the reference parameters file
                    compare_files_without_order('OBC_parameters.txt',
                                                reference_file)

        # Test for ethylene
        test_OBCOPLS_writer_ligand(pdbfile='ligands/ethylene.pdb',
                                   tag_name='ETL',
                                   ffld_name='tests/ETL_ffld_output.txt',
                                   reference_file=TEMPLATE_PARAMS_ETL)

        # Test for methane
        test_OBCOPLS_writer_ligand(pdbfile='ligands/methane.pdb',
                                   tag_name='MET',
                                   ffld_name='tests/MET_ffld_output.txt',
                                   reference_file=TEMPLATE_PARAMS_MET)

        # Test for malonate
        test_OBCOPLS_writer_ligand(pdbfile='ligands/malonate.pdb',
                                   tag_name='MAL',
                                   ffld_name='tests/MAL_ffld_output.txt',
                                   reference_file=TEMPLATE_PARAMS_MAL)



"""
This module contains the tests to check some handy classes and functions
of peleffy.
"""


import pytest

import io
import os
import tempfile

from peleffy.utils import Logger
from peleffy.topology import Molecule


class TestLogger(object):
    def test_logger_levels(self):
        """
        It checks the correct behaviour of the different log levels.
        """
        def push_messages(log):
            """Pull some messages at different levels."""
            log.debug('Debug message')
            log.info('Info message')
            log.warning('Warn message')
            log.error('Error message')
            log.critical('Critical message')

        import logging

        # Force a hard reset of logging library and the logger it manages
        from importlib import reload
        logging.shutdown()
        reload(logging)

        # Initiate logger
        log = Logger()

        # Try the default level (INFO)
        # Catch logger messages to string buffer
        with io.StringIO() as buf:
            # Add custom handler to logger
            log_handler = logging.StreamHandler(buf)
            log._logger.handlers = list()
            log._logger.addHandler(log_handler)

            # Push messages
            push_messages(log)

            # Get string from buffer
            output = buf.getvalue()

            assert output == 'Info message\nWarn message\n' \
                + 'Error message\nCritical message\n', \
                'Unexpected logger message at standard output'

        # Try DEBUG level
        # Catch logger messages to string buffer
        with io.StringIO() as buf:
            # Add custom handler to logger
            log_handler = logging.StreamHandler(buf)
            log._logger.handlers = list()
            log._logger.addHandler(log_handler)

            # Try DEBUG level
            log.set_level('DEBUG')

            # Push messages
            push_messages(log)

            # Get string from buffer
            output = buf.getvalue()

            assert output == 'Debug message\nInfo message\n'\
                + 'Warn message\nError message\nCritical message\n', \
                'Unexpected logger message at standard output'

        # Try INFO level
        # Catch logger messages to string buffer
        with io.StringIO() as buf:
            # Add custom handler to logger
            log_handler = logging.StreamHandler(buf)
            log._logger.handlers = list()
            log._logger.addHandler(log_handler)

            # Try INFO level
            log.set_level('INFO')

            # Push messages
            push_messages(log)

            # Get string from buffer
            output = buf.getvalue()

            assert output == 'Info message\nWarn message\n' \
                + 'Error message\nCritical message\n', \
                'Unexpected logger message at standard output'

        # Try WARNING level
        # Catch logger messages to string buffer
        with io.StringIO() as buf:
            # Add custom handler to logger
            log_handler = logging.StreamHandler(buf)
            log._logger.handlers = list()
            log._logger.addHandler(log_handler)

            # Try WARNING level
            log.set_level('WARNING')

            # Push messages
            push_messages(log)

            # Get string from buffer
            output = buf.getvalue()

            assert output == 'Warn message\nError message\n' \
                + 'Critical message\n', \
                'Unexpected logger message at standard output'

        # Try ERROR level
        # Catch logger messages to string buffer
        with io.StringIO() as buf:
            # Add custom handler to logger
            log_handler = logging.StreamHandler(buf)
            log._logger.handlers = list()
            log._logger.addHandler(log_handler)

            # Try ERROR level
            log.set_level('ERROR')

            # Push messages
            push_messages(log)

            # Get string from buffer
            output = buf.getvalue()

            assert output == 'Error message\nCritical message\n', \
                'Unexpected logger message at standard output'

        # Try CRITICAL level
        # Catch logger messages to string buffer
        with io.StringIO() as buf:
            # Add custom handler to logger
            log_handler = logging.StreamHandler(buf)
            log._logger.handlers = list()
            log._logger.addHandler(log_handler)

            # Try CRITICAL level
            log.set_level('CRITICAL')

            # Push messages
            push_messages(log)

            # Get string from buffer
            output = buf.getvalue()

            assert output == 'Critical message\n', \
                'Unexpected logger message at standard output'


class TestOutputPathHandler(object):
    """
    It contains all the tests to validate the OuputPathHandler class.
    """

    def test_non_datalocal_paths(self):
        """
        It tests the non-datalocal paths assignment.
        """
        from peleffy.utils import OutputPathHandler
        from peleffy.forcefield import OpenForceField

        # Load benzene molecule
        molecule = Molecule(smiles='c1ccccc1', name='benzene', tag='BNZ')

        # Load force field
        openff = OpenForceField('openff_unconstrained-1.2.1.offxml')

        # Molecule's tag
        tag = molecule.tag

        # Initialize output handler without output_path
        output_handler = OutputPathHandler(molecule, openff,
                                           as_datalocal=False)

        # Validate output paths
        assert output_handler.get_rotamer_library_path() == \
            './{}.rot.assign'.format(tag.upper()), \
            'Unexpected default rotamer library path'
        assert output_handler.get_impact_template_path() == \
            './{}z'.format(tag.lower()), \
            'Unexpected default Impact template path'
        assert output_handler.get_solvent_template_path() == \
            './ligandParams.txt', \
            'Unexpected default solvent parameters path'
        assert output_handler.get_conformation_library_path() == \
            './{}.conformation'.format(tag.upper()), \
            'Unexpected default conformation library path'

        # Initialize output handler with an output_path set
        with tempfile.TemporaryDirectory() as tmpdir:
            output_handler = OutputPathHandler(
                molecule, openff, as_datalocal=False,
                output_path=os.path.join(tmpdir, 'output'))

            assert output_handler.get_rotamer_library_path() == \
                os.path.join(tmpdir, 'output',
                             '{}.rot.assign'.format(tag.upper())), \
                'Unexpected default rotamer library path'
            assert output_handler.get_impact_template_path() == \
                os.path.join(tmpdir, 'output', '{}z'.format(tag.lower())), \
                'Unexpected default Impact template path'
            assert output_handler.get_solvent_template_path() == \
                os.path.join(tmpdir, 'output', 'ligandParams.txt'), \
                'Unexpected default solvent parameters path'
            assert output_handler.get_conformation_library_path() == \
                os.path.join(tmpdir, 'output',
                             '{}.conformation'.format(tag.upper())), \
                'Unexpected default conformation library path'

    def test_datalocal_paths_for_openff(self):
        """It tests the datalocal paths assignment for OpenFF."""
        from peleffy.utils import OutputPathHandler
        from peleffy.forcefield import OpenForceField

        # Load benzene molecule
        molecule = Molecule(smiles='c1ccccc1', name='benzene', tag='BNZ')

        # Load force field
        openff = OpenForceField('openff_unconstrained-1.2.1.offxml')
        # Molecule's tag
        tag = molecule.tag

        # Initialize output handler without output_path
        output_handler = OutputPathHandler(molecule, openff,
                                           as_datalocal=True)

        # Validate output paths
        assert output_handler.get_rotamer_library_path(
            create_missing_folders=False) == \
            './DataLocal/LigandRotamerLibs/' \
            + '{}.rot.assign'.format(tag.upper()), \
            'Unexpected default rotamer library path'
        assert output_handler.get_impact_template_path(
            create_missing_folders=False) == \
            './DataLocal/Templates/OpenFF/Parsley/' \
            + '{}z'.format(tag.lower()), \
            'Unexpected default Impact template path'
        assert output_handler.get_solvent_template_path(
            create_missing_folders=False) == \
            './DataLocal/OBC/ligandParams.txt', \
            'Unexpected default solvent parameters path'
        assert output_handler.get_conformation_library_path(
            create_missing_folders=False) == \
            './DataLocal/Conformations/' \
            + '{}.conformation'.format(tag.upper()), \
            'Unexpected default conformation library path'

        with tempfile.TemporaryDirectory() as tmpdir:
            output_handler = OutputPathHandler(
                molecule, openff, as_datalocal=True,
                output_path=os.path.join(tmpdir, 'output'))

            # Validate output paths
            assert output_handler.get_rotamer_library_path(
                create_missing_folders=False) == \
                tmpdir + '/output/DataLocal/LigandRotamerLibs/' \
                + '{}.rot.assign'.format(tag.upper()), \
                'Unexpected default rotamer library path'
            assert output_handler.get_impact_template_path(
                create_missing_folders=False) == \
                tmpdir + '/output/DataLocal/Templates/OpenFF/Parsley/' \
                + '{}z'.format(tag.lower()), \
                'Unexpected default Impact template path'
            assert output_handler.get_solvent_template_path(
                create_missing_folders=False) == \
                tmpdir + '/output/DataLocal/OBC/ligandParams.txt', \
                'Unexpected default solvent parameters path'
            assert output_handler.get_conformation_library_path(
                create_missing_folders=False) == \
                tmpdir + '/output/DataLocal/Conformations/' \
                + '{}.conformation'.format(tag.upper()), \
                'Unexpected default conformation library path'

    def test_datalocal_paths_for_opls(self):
        """It tests the datalocal paths assignment for OPLS2005."""
        from peleffy.utils import OutputPathHandler
        from peleffy.forcefield import OPLS2005ForceField

        # Load benzene molecule
        molecule = Molecule(smiles='c1ccccc1', name='benzene', tag='BNZ')

        # Load force field
        opls2005 = OPLS2005ForceField()

        # Molecule's tag
        tag = molecule.tag

        # Initialize output handler without output_path
        output_handler = OutputPathHandler(molecule, opls2005,
                                           as_datalocal=True)

        # Validate output paths
        assert output_handler.get_rotamer_library_path(
            create_missing_folders=False) == \
            './DataLocal/LigandRotamerLibs/' \
            + '{}.rot.assign'.format(tag.upper()), \
            'Unexpected default rotamer library path'
        assert output_handler.get_impact_template_path(
            create_missing_folders=False) == \
            './DataLocal/Templates/OPLS2005/HeteroAtoms/' \
            + '{}z'.format(tag.lower()), \
            'Unexpected default Impact template path'
        assert output_handler.get_solvent_template_path(
            create_missing_folders=False) == \
            './DataLocal/OBC/ligandParams.txt', \
            'Unexpected default solvent parameters path'
        assert output_handler.get_conformation_library_path(
            create_missing_folders=False) == \
            './DataLocal/Conformations/' \
            + '{}.conformation'.format(tag.upper()), \
            'Unexpected default conformation library path'

        # Initialize output handler with an output_path set
        with tempfile.TemporaryDirectory() as tmpdir:
            output_handler = OutputPathHandler(
                molecule, opls2005, as_datalocal=True,
                output_path=os.path.join(tmpdir, 'output'))

            assert output_handler.get_rotamer_library_path(
                create_missing_folders=False) == \
                os.path.join(tmpdir, 'output', 'DataLocal/LigandRotamerLibs/'
                             + '{}.rot.assign'.format(tag.upper())), \
                'Unexpected default rotamer library path'
            assert output_handler.get_impact_template_path(
                create_missing_folders=False) == \
                os.path.join(tmpdir, 'output', 'DataLocal/Templates/OPLS2005/'
                             + 'HeteroAtoms/{}z'.format(tag.lower())), \
                'Unexpected default Impact template path'
            assert output_handler.get_solvent_template_path(
                create_missing_folders=False) == \
                os.path.join(tmpdir, 'output',
                             'DataLocal/OBC/ligandParams.txt'), \
                'Unexpected default solvent parameters path'
            assert output_handler.get_conformation_library_path(
                create_missing_folders=False) == \
                os.path.join(tmpdir, 'output/DataLocal/Conformations/' \
                + '{}.conformation'.format(tag.upper())), \
                'Unexpected default conformation library path'

    def test_datalocal_paths_for_offopls(self):
        """
        It tests the datalocal paths assignment for OpenFF-OPLS2005
        force field.
        """
        from peleffy.utils import OutputPathHandler
        from peleffy.forcefield import OpenFFOPLS2005ForceField

        # Load benzene molecule
        molecule = Molecule(smiles='c1ccccc1', name='benzene', tag='BNZ')
        molecule._forcefield = OpenFFOPLS2005ForceField('OPLS2005')

        # Load force field
        hybridff = OpenFFOPLS2005ForceField(
            'openff_unconstrained-1.2.1.offxml')

        # Molecule's tag
        tag = molecule.tag

        # Initialize output handler without output_path
        output_handler = OutputPathHandler(molecule, hybridff,
                                           as_datalocal=True)

        # Validate output paths
        assert output_handler.get_rotamer_library_path(
            create_missing_folders=False) == \
            './DataLocal/LigandRotamerLibs/' \
            + '{}.rot.assign'.format(tag.upper()), \
            'Unexpected default rotamer library path'
        assert output_handler.get_impact_template_path(
            create_missing_folders=False) == \
            './DataLocal/Templates/OpenFF/Parsley/' \
            + '{}z'.format(tag.lower()), \
            'Unexpected default Impact template path'
        assert output_handler.get_solvent_template_path(
            create_missing_folders=False) == \
            './DataLocal/OBC/ligandParams.txt', \
            'Unexpected default solvent parameters path'
        assert output_handler.get_conformation_library_path(
            create_missing_folders=False) == \
            './DataLocal/Conformations/' \
            + '{}.conformation'.format(tag.upper()), \
            'Unexpected default conformation library path'

        # Initialize output handler with an output_path set
        with tempfile.TemporaryDirectory() as tmpdir:
            output_handler = OutputPathHandler(
                molecule, hybridff, as_datalocal=True,
                output_path=os.path.join(tmpdir, 'output'))

            assert output_handler.get_rotamer_library_path(
                create_missing_folders=False) == \
                os.path.join(tmpdir, 'output', 'DataLocal/LigandRotamerLibs/'
                             + '{}.rot.assign'.format(tag.upper())), \
                'Unexpected default rotamer library path'
            assert output_handler.get_impact_template_path(
                create_missing_folders=False) == \
                os.path.join(tmpdir, 'output', 'DataLocal/Templates/'
                             + 'OpenFF/Parsley/{}z'.format(tag.lower())), \
                'Unexpected default Impact template path'
            assert output_handler.get_solvent_template_path(
                create_missing_folders=False) == \
                os.path.join(tmpdir, 'output',
                             'DataLocal/OBC/ligandParams.txt'), \
                'Unexpected default solvent parameters path'
            assert output_handler.get_conformation_library_path(
                create_missing_folders=False) == \
                os.path.join(tmpdir, 'output/DataLocal/Conformations/' \
                + '{}.conformation'.format(tag.upper())), \
                'Unexpected default conformation library path'

        # Initialize output handler without output_path
        output_handler = OutputPathHandler(molecule, hybridff,
                                           as_datalocal=True)

        # Set force field source for nonbonding parameters
        hybridff.set_nonbonding_parameters('opls2005')

        # Validate output paths
        assert output_handler.get_rotamer_library_path(
            create_missing_folders=False) == \
            './DataLocal/LigandRotamerLibs/' \
            + '{}.rot.assign'.format(tag.upper()), \
            'Unexpected default rotamer library path'
        assert output_handler.get_impact_template_path(
            create_missing_folders=False) == \
            './DataLocal/Templates/OPLS2005/HeteroAtoms/' \
            + '{}z'.format(tag.lower()), \
            'Unexpected default Impact template path'
        assert output_handler.get_solvent_template_path(
            create_missing_folders=False) == \
            './DataLocal/OBC/ligandParams.txt', \
            'Unexpected default solvent parameters path'
        assert output_handler.get_conformation_library_path(
            create_missing_folders=False) == \
            './DataLocal/Conformations/' \
            + '{}.conformation'.format(tag.upper()), \
            'Unexpected default conformation library path'

        # Initialize output handler with an output_path set
        with tempfile.TemporaryDirectory() as tmpdir:
            output_handler = OutputPathHandler(
                molecule, hybridff, as_datalocal=True,
                output_path=os.path.join(tmpdir, 'output'))

            assert output_handler.get_rotamer_library_path(
                create_missing_folders=False) == \
                os.path.join(tmpdir, 'output', 'DataLocal/LigandRotamerLibs/'
                             + '{}.rot.assign'.format(tag.upper())), \
                'Unexpected default rotamer library path'
            assert output_handler.get_impact_template_path(
                create_missing_folders=False) == \
                os.path.join(tmpdir, 'output', 'DataLocal/Templates/OPLS2005/'
                             + 'HeteroAtoms/{}z'.format(tag.lower())), \
                'Unexpected default Impact template path'
            assert output_handler.get_solvent_template_path(
                create_missing_folders=False) == \
                os.path.join(tmpdir, 'output',
                             'DataLocal/OBC/ligandParams.txt'), \
                'Unexpected default solvent parameters path'
            assert output_handler.get_conformation_library_path(
                create_missing_folders=False) == \
                os.path.join(tmpdir, 'output',
                             'DataLocal/Conformations/',
                             '{}.conformation'.format(tag.upper())), \
                'Unexpected default conformation library path'

    def test_folder_creation(self):
        """
        It tests the folder creation of the OutputPathHandler class.
        """
        from peleffy.utils import OutputPathHandler, temporary_cd
        from peleffy.forcefield import OpenForceField

        # Load benzene molecule
        molecule = Molecule(smiles='c1ccccc1', name='benzene', tag='BNZ')

        # Load force field
        openff = OpenForceField('openff_unconstrained-1.2.1.offxml')

        # Initialize output handler with an output_path set
        with tempfile.TemporaryDirectory() as tmpdir:
            # Initialize output handler without output_path
            output_handler = OutputPathHandler(
                molecule, openff, as_datalocal=True,
                output_path=os.path.join(tmpdir, 'output'))

            # Test path getter without folder creation
            path = output_handler.get_rotamer_library_path(
                create_missing_folders=False)
            path_dir = os.path.dirname(path)
            assert os.path.isdir(path_dir) is False, \
                'This directory should not exist'
            path = output_handler.get_impact_template_path(
                create_missing_folders=False)
            path_dir = os.path.dirname(path)
            assert os.path.isdir(path_dir) is False, \
                'This directory should not exist'
            path = output_handler.get_solvent_template_path(
                create_missing_folders=False)
            path_dir = os.path.dirname(path)
            assert os.path.isdir(path_dir) is False, \
                'This directory should not exist'
            path = output_handler.get_conformation_library_path(
                create_missing_folders=False)
            path_dir = os.path.dirname(path)
            assert os.path.isdir(path_dir) is False, \
                'This directory should not exist'

        # Initialize output handler with an output_path set
        with tempfile.TemporaryDirectory() as tmpdir:
            with temporary_cd(tmpdir):
                # Initialize output handler without output_path
                output_handler = OutputPathHandler(
                    molecule, openff, as_datalocal=True,
                    output_path=os.path.join(tmpdir, 'output'))

                # Test path getter with folder creation
                path = output_handler.get_rotamer_library_path(
                    create_missing_folders=True)
                path_dir = os.path.dirname(path)
                assert os.path.isdir(path_dir) is True, \
                    'This directory should exist'
                path = output_handler.get_impact_template_path(
                    create_missing_folders=True)
                path_dir = os.path.dirname(path)
                assert os.path.isdir(path_dir) is True, \
                    'This directory should exist'
                path = output_handler.get_solvent_template_path(
                    create_missing_folders=True)
                path_dir = os.path.dirname(path)
                assert os.path.isdir(path_dir) is True, \
                    'This directory should exist'
                path = output_handler.get_conformation_library_path(
                    create_missing_folders=True)
                path_dir = os.path.dirname(path)
                assert os.path.isdir(path_dir) is True, \
                    'This directory should exist'


class TestMAEParser(object):
    """
    It contains all the tests to validate functions that parse MAE files.
    """

    def test_parse_charges_from_mae(self):
        """
        It tests the MAE parser for assigning partial charges from an external
        file.
        """
        from peleffy.utils import get_data_file_path, parse_charges_from_mae
        from peleffy.topology import Molecule
        from peleffy.forcefield import OpenForceField

        FORCEFIELD = 'openff_unconstrained-1.2.0.offxml'

        PATH_MAE_MAT = get_data_file_path('tests/MAT.mae')
        PATH_PDB_MAT = get_data_file_path('tests/MAT.pdb')
        PATH_MAE_BHP = get_data_file_path('ligands/BHP.mae')
        PATH_PDB_BHP = get_data_file_path('ligands/BHP.pdb')
        PATH_MAE_ETL = get_data_file_path('ligands/ethylene.mae')
        PATH_PDB_ETL = get_data_file_path('ligands/ethylene.pdb')
        PATH_PDB_MAL = get_data_file_path('ligands/malonate.pdb')

        CHARGES_REFERENCE_BHP = [[-0.35703, 'O1'],
                                 [-0.59535, 'O2'],
                                 [-0.50292, 'O3'],
                                 [-0.25243, 'C1'],
                                 [0.30438, 'C2'],
                                 [0.22092, 'C3'],
                                 [0.54336, 'C4'],
                                 [-0.20569, 'C5'],
                                 [-0.20192, 'C6'],
                                 [0.16631, 'C7'],
                                 [0.02422, 'C8'],
                                 [-0.09115, 'C9'],
                                 [-0.09904, 'C10'],
                                 [-0.15673, 'C11'],
                                 [-0.13245, 'C12'],
                                 [-0.17806, 'C13'],
                                 [-0.12489, 'C14'],
                                 [-0.09307, 'C15'],
                                 [-0.08973, 'C16'],
                                 [0.05397, 'H1'],
                                 [0.07338, 'H2'],
                                 [0.04514, 'H3'],
                                 [0.12979, 'H4'],
                                 [0.11025, 'H5'],
                                 [0.04054, 'H6'],
                                 [0.04581, 'H7'],
                                 [0.11444, 'H8'],
                                 [0.11761, 'H9'],
                                 [0.37274, 'H10'],
                                 [0.11825, 'H11'],
                                 [0.12584, 'H12'],
                                 [0.1381, 'H13'],
                                 [0.11696, 'H14'],
                                 [0.11148, 'H15'],
                                 [0.10697, 'H16']]

        CHARGES_REFERENCE_MAT = [[-0.18938, 'F1'],
                                 [-0.21715, 'F2'],
                                 [-0.21234, 'F3'],
                                 [-0.39736, 'O1'],
                                 [-0.58890, 'O2'],
                                 [-1.00825, 'N1'],
                                 [0.72066, 'C1'],
                                 [-0.06281, 'C2'],
                                 [-0.67474, 'C3'],
                                 [0.10391, 'C4'],
                                 [0.16293, 'C5'],
                                 [-0.61076, 'C6'],
                                 [0.78183, 'C7'],
                                 [0.27041, 'C8'],
                                 [-0.48769, 'C9'],
                                 [0.15704, 'C10'],
                                 [-0.02646, 'H1'],
                                 [-0.08394, 'H2'],
                                 [-0.01308, 'H3'],
                                 [0.14006, 'H4'],
                                 [0.12960, 'H5'],
                                 [0.31245, 'H6'],
                                 [0.30268, 'H7'],
                                 [0.17026, 'H8'],
                                 [0.15782, 'H9'],
                                 [0.03175, 'C11'],
                                 [0.03894, 'H10'],
                                 [0.07509, 'H11'],
                                 [0.01743, 'H12']]

        # Check up correct charges for malonate
        m = Molecule(PATH_PDB_MAT)
        openff = OpenForceField(FORCEFIELD)
        parameters = openff.parameterize(m, charge_method='dummy')
        parameters = parse_charges_from_mae(PATH_MAE_MAT, parameters)

        for charge, atom_name in zip(parameters['charges'],
                                     parameters['atom_names']):
            assert [charge._value, atom_name.replace('_', '')] in \
                CHARGES_REFERENCE_MAT, \
                'Incorrect charge value for {}.'.format(atom_name)

        # Check up correct charges for BHP
        m = Molecule(PATH_PDB_BHP)
        openff = OpenForceField(FORCEFIELD)
        parameters = openff.parameterize(m, charge_method='dummy')
        parameters = parse_charges_from_mae(PATH_MAE_BHP, parameters)

        for charge, atom_name in zip(parameters['charges'],
                                     parameters['atom_names']):
            assert [charge._value, atom_name.replace('_', '')] in \
                CHARGES_REFERENCE_BHP, \
                'Incorrect charge value for {}.'.format(atom_name)

        # Error: MAE file without charges information
        m = Molecule(PATH_PDB_ETL)
        openff = OpenForceField(FORCEFIELD)
        parameters = openff.parameterize(m, charge_method='dummy')
        with pytest.raises(ValueError):
            _ = parse_charges_from_mae(PATH_MAE_ETL, parameters)

        # Error: Inconsistency between Moelcule atom names and MAE atom names
        m = Molecule(PATH_PDB_MAL)
        openff = OpenForceField(FORCEFIELD)
        parameters = openff.parameterize(m, charge_method='dummy')
        with pytest.raises(ValueError):
            _ = parse_charges_from_mae(PATH_PDB_BHP, parameters)


class TestPDBFile(object):
    """
    It contains all the tests to validate the PDBFile class.
    """
    def test_get_molecule_from_chain(self):
        """
        It tests the method that returns a molecule from a selected chain from
        an input PDB file.
        """
        def compare_molecules(mol1, mol2):
            """
            It compares that two molecules contain the same PDB atom names and
            that have been initialized with the same flags.

            Parameters
            ----------
            mol1 : a peleffy.topology.Molecule object
                Molecule.
            mol2 : a peleffy.topology.Molecule object
                Reference molecule.

            Returns
            -------
            are_equal : bool
                Wheter the molecules are equal or not.
            """
            if mol1._rotamer_resolution == mol2._rotamer_resolution \
                    and mol1._exclude_terminal_rotamers ==  \
            mol2._exclude_terminal_rotamers\
                    and mol1._allow_undefined_stereo == \
            mol2._allow_undefined_stereo\
                    and mol1.get_pdb_atom_names() == \
            mol2.get_pdb_atom_names():
                return True
            else:
                return False

        from peleffy.utils.input import PDBFile
        from peleffy.utils import get_data_file_path
        from peleffy.topology import Molecule

        PATH_COMPLEX_PDB = get_data_file_path('complexes/LYS_BNZ.pdb')
        PATH_LIGAND_PDB = get_data_file_path('ligands/BNZ.pdb')

        # Test method get_molecule_from_chain
        PDBreader = PDBFile(PATH_COMPLEX_PDB)
        molecules = PDBreader.get_molecules_from_chain(selected_chain='L')
        assert len(molecules) == 1, 'Unexpected number of molecules'
        assert compare_molecules(molecules[0], Molecule(PATH_LIGAND_PDB)) is True

        # Test allow_undefined_stereo flag
        PDBreader = PDBFile(PATH_COMPLEX_PDB)
        molecules = \
            PDBreader.get_molecules_from_chain(selected_chain='L',
                                               allow_undefined_stereo=True)
        assert len(molecules) == 1, 'Unexpected number of molecules'
        assert compare_molecules(molecules[0],
                                 Molecule(PATH_LIGAND_PDB,
                                          allow_undefined_stereo=True)) is True

        # Test exclude_terminal_rotamers flag
        PDBreader = PDBFile(PATH_COMPLEX_PDB)
        molecules = \
            PDBreader.get_molecules_from_chain(selected_chain='L',
                                               exclude_terminal_rotamers=False)
        assert len(molecules) == 1, 'Unexpected number of molecules'
        assert compare_molecules(molecules[0], Molecule(
            PATH_LIGAND_PDB, exclude_terminal_rotamers=False)) is True

        # Test rotamer_resolution flag
        PDBreader = PDBFile(PATH_COMPLEX_PDB)
        molecules = PDBreader.get_molecules_from_chain(selected_chain='L',
                                                       rotamer_resolution=10)
        assert compare_molecules(molecules[0],
                                 Molecule(PATH_LIGAND_PDB,
                                          rotamer_resolution=10)) is True

        # Test core_constraints flag
        PDBreader = PDBFile(PATH_COMPLEX_PDB)
        molecules = PDBreader.get_molecules_from_chain(selected_chain='L',
                                                       core_constraints=[1,2])
        assert len(molecules) == 1, 'Unexpected number of molecules'
        assert molecules[0].core_constraints == [1,2]


    def test_get_hetero_molecules(self):
        """
        It tests the method that returns a list with all the hetero molecules in
        an input PDB.
        """
        def compare_molecules(mol1, mol2):
            """
            It compares that two molecules contain the same PDB atom names and
            that have been initialized with the same flags.

            Parameters
            ----------
            mol1 : a peleffy.topology.Molecule object
                Molecule.
            mol2 : a peleffy.topology.Molecule object
                Reference molecule.

            Returns
            -------
            are_equal : bool
                Wheter the molecules are equal or not.
            """
            if mol1._rotamer_resolution == mol2._rotamer_resolution \
                    and mol1._exclude_terminal_rotamers ==  \
            mol2._exclude_terminal_rotamers\
                    and mol1._allow_undefined_stereo == \
            mol2._allow_undefined_stereo\
                    and mol1.get_pdb_atom_names() == \
            mol2.get_pdb_atom_names():
                return True
            else:
                return False

        from peleffy.utils.input import PDBFile
        from peleffy.utils import get_data_file_path
        from peleffy.topology import Molecule

        PATH_COMPLEX_PDB = get_data_file_path('complexes/LIG1_LIG2.pdb')
        PATH_LIGAND1 = get_data_file_path('ligands/LIG1.pdb')
        PATH_LIGAND2 = get_data_file_path('ligands/LIG2.pdb')

        # Test extract multiple molecules
        PDBreader = PDBFile(PATH_COMPLEX_PDB)
        molecules = PDBreader.get_hetero_molecules()

        for molecule in molecules:
            for molecule_ref in [Molecule(PATH_LIGAND1),Molecule(PATH_LIGAND2)]:
                if molecule._tag == molecule_ref.tag:
                    assert compare_molecules(molecule, molecule_ref)

        # Test set ligand core constraints
        PATH_COMPLEX_PDB = get_data_file_path('complexes/complex_test.pdb')
        molecules = \
            PDBreader.get_hetero_molecules(allow_undefined_stereo = True,
                                           ligand_core_constraints = [1,2,3],
                                           ligand_resname = 'ANL')
        for molecule in molecules:
            if molecule._tag == 'ANL':
                assert molecule.core_constraints == [1,2,3]
            else:
                assert molecule.core_constraints == []


    def test_raise_errors(self):
        """
        It tests all the possible errors when geting a molecule from a chain.
        """
        from peleffy.utils.input import PDBFile
        from peleffy.utils import get_data_file_path

        PATH_COMPLEX_PDB = get_data_file_path('complexes/LYS_BNZ.pdb')

        # The chain selected does not exist in the PDB file
        with pytest.raises(ValueError):
            PDBreader = PDBFile(PATH_COMPLEX_PDB)
            _ = PDBreader.get_molecules_from_chain(selected_chain='F')

        # The chain selected is not an hetero molecule
        with pytest.raises(ValueError):
            PDBreader = PDBFile(PATH_COMPLEX_PDB)
            _ = PDBreader.get_molecules_from_chain(selected_chain='A')



# -*- coding: utf-8 -*-
"""
This module is designed to run peleffy through the command-line.
"""

__author__ = "Marti Municoy"
__license__ = "GPL"
__maintainer__ = "Marti Municoy"
__email__ = "marti.municoy@bsc.es"


import os
import argparse as ap

import peleffy
from peleffy.utils import Logger, OutputPathHandler
from peleffy.forcefield.selectors import ChargeCalculatorSelector


DEFAULT_OFF_FORCEFIELD = 'openff_unconstrained-1.3.0.offxml'
DEFAULT_RESOLUTION = int(30)
DEFAULT_CHARGE_METHOD = 'am1bcc'
AVAILABLE_CHARGE_METHODS = ChargeCalculatorSelector()._AVAILABLE_TYPES.keys()
IMPACT_TEMPLATE_PATH = 'DataLocal/Templates/OFF/Parsley/HeteroAtoms/'
ROTAMER_LIBRARY_PATH = 'DataLocal/LigandRotamerLibs/'
SOLVENT_TEMPLATE_PATH = 'DataLocal/OBC/'
DEFAULT_TERMINAL_ROT_TO_IGNORE = 1


def parse_args(args):
    """
    It parses the command-line arguments.

    Parameters
    ----------
    args : list[str]
        List of command-line arguments to parse

    Returns
    -------
    parsed_args : argparse.Namespace
        It contains the command-line arguments that are supplied by the user
    """
    parser = ap.ArgumentParser()
    parser.add_argument("pdb_file", metavar="PDB FILE", type=str,
                        help="Path PDB file to parameterize")
    parser.add_argument("-f", "--forcefield", metavar="NAME",
                        type=str, help="OpenForceField's forcefield name. "
                        + "Default is " + str(DEFAULT_OFF_FORCEFIELD),
                        default=DEFAULT_OFF_FORCEFIELD)
    parser.add_argument("-r", "--resolution", metavar="INT", type=int,
                        help="Rotamer library resolution in degrees. "
                        + "Default is " + str(DEFAULT_RESOLUTION),
                        default=DEFAULT_RESOLUTION)
    parser.add_argument("-o", "--output", metavar="PATH",
                        help="Output path. Default is the current working "
                        + "directory")
    parser.add_argument("--conformations_info_path", default=None, type=str,
                        help="Path to the folder containing the BCE output"
                        + " used to collect conformations for PELE")
    parser.add_argument('--with_solvent', dest='with_solvent',
                        help="Generate solvent parameters for OBC",
                        action='store_true')
    parser.add_argument('--as_datalocal', dest='as_datalocal',
                        help="Output will be saved following PELE's DataLocal "
                        + "hierarchy", action='store_true')
    parser.add_argument('-c', '--charge_method', metavar="NAME",
                        type=str, help="The name of the method to use to "
                        + "compute charges", default=DEFAULT_CHARGE_METHOD,
                        choices=AVAILABLE_CHARGE_METHODS)
    parser.add_argument('--charges_from_file', metavar="PATH",
                        type=str, help="The path to the file with charges",
                        default=None)
    parser.add_argument("--chain", metavar="CHAIN",
                        type=str, help="Chain of the molecule to parameterize",
                        default=None)
    parser.add_argument('--include_terminal_rotamers',
                        dest="include_terminal_rotamers",
                        action='store_true',
                        help="Not exclude terminal rotamers "
                        + "when building the rotamer library")
    parser.add_argument('-s', '--silent',
                        dest="silent",
                        action='store_true',
                        help="Activate silent mode")
    parser.add_argument('-d', '--debug',
                        dest="debug",
                        action='store_true',
                        help="Activate debug mode")

    parser.set_defaults(as_datalocal=False)
    parser.set_defaults(with_solvent=False)
    parser.set_defaults(include_terminal_rotamers=False)
    parser.set_defaults(silent=False)
    parser.set_defaults(debug=False)

    parsed_args = parser.parse_args(args)

    return parsed_args


def run_peleffy(pdb_file,
                forcefield_name=DEFAULT_OFF_FORCEFIELD,
                resolution=DEFAULT_RESOLUTION,
                charge_method=DEFAULT_CHARGE_METHOD,
                charges_from_file=None,
                chain=None,
                exclude_terminal_rotamers=True,
                output=None, with_solvent=False, as_datalocal=False,
                conformation_path=None):
    """
    It runs peleffy.

    Parameters
    ----------
    pdb_file : str
        The path to the pdb_file to parameterize with peleffy
    forcefield_name : str
        The name of an OpenForceField's forcefield
    resolution : float
        The resolution in degrees for the rotamer library. Default is 30
    charge_method : str
        The name of the method to use to compute partial charges. Default
        is 'am1bcc'
    charges_from_file : str
        The file containing the partial charges to assign to the
        molecule. Default is None
    chain : str
        Chain to the molecule if the PDB contains multiple molecules.
    exclude_terminal_rotamers : bool
        Whether to exclude terminal rotamers or not
    output : str
        Path where output files will be saved
    with_solvent : bool
        Whether to generate and save the solvent parameters for the input
        molecule or not
    as_datalocal : bool
        Whether to save output files following PELE's DataLocal hierarchy or
        not
    conformation_path: str
        Path to the BCE server outupt to use to extract dihedral angles
    dihedral_mode: str
        Select what kind of dihedrals to extract (all or only flexible)
    """
    if charges_from_file is not None:
        charge_method_str = 'file\n' \
            + '   - Charge file: {}'.format(charges_from_file)
        charge_method = 'dummy'
    else:
        charge_method_str = charge_method

    log = Logger()
    log.info('-' * 60)
    log.info('Open Force Field parameterizer for PELE', peleffy.__version__)
    log.info('-' * 60)
    log.info(' - General:')
    log.info('   - Input PDB:', pdb_file)
    log.info('   - Output path:', output)
    log.info('   - Write solvent parameters:', with_solvent)
    log.info('   - DataLocal-like output:', as_datalocal)
    log.info(' - Parameterization:')
    log.info('   - Force field:', forcefield_name)
    log.info('   - Charge method:', charge_method_str)
    log.info(' - Rotamer library:')
    log.info('   - Resolution:', resolution)
    log.info('   - Exclude terminal rotamers:', exclude_terminal_rotamers)
    log.info('-' * 60)

    from peleffy.topology import Molecule, BCEConformations
    from peleffy.template import Impact
    from peleffy.solvent import OBC2
    from peleffy.forcefield import ForceFieldSelector
    from peleffy.topology import Topology
    from peleffy.utils import parse_charges_from_mae
    from peleffy.utils.input import PDBFile

    if not output:
        output = os.getcwd()

    # Initialize molecule
    if chain is not None:
        PDBreader = PDBFile(pdb_file)
        molecule = PDBreader.get_molecules_from_chain(
            selected_chain=chain,
            rotamer_resolution=resolution,
            exclude_terminal_rotamers=exclude_terminal_rotamers)
    else:
        molecule = Molecule(
            pdb_file, rotamer_resolution=resolution,
            exclude_terminal_rotamers=exclude_terminal_rotamers)

    # Initialize force field
    ff_selector = ForceFieldSelector()
    forcefield = ff_selector.get_by_name(forcefield_name)

    output_handler = OutputPathHandler(molecule, forcefield,
                                       output_path=output,
                                       as_datalocal=as_datalocal)

    # if conformation_path is set, we don't want a rotamer library
    if conformation_path is None:
        rotamer_library = peleffy.topology.RotamerLibrary(molecule)
        rotamer_library.to_file(output_handler.get_rotamer_library_path())

    # Parameterize molecule with the selected force field
    log.info(' - Parameterizing molecule')
    parameters = forcefield.parameterize(molecule,
                                         charge_method=charge_method)

    # Update charge parameters from the MAE file
    if charges_from_file is not None:
        parameters = parse_charges_from_mae(charges_from_file, parameters)

    # Generate the molecular topology
    topology = Topology(molecule, parameters)
    log.info(' - Parameters were built successfully:')
    log.info('   - {} atoms'.format(len(topology.atoms)))
    log.info('   - {} bonds'.format(len(topology.bonds)))
    log.info('   - {} torsions'.format(len(topology.angles)))
    log.info('   - {} propers'.format(len(topology.propers)))
    log.info('   - {} impropers'.format(len(topology.impropers)))

    # Generate the impact template
    impact = Impact(topology)
    impact.to_file(output_handler.get_impact_template_path())

    # Generate the solvent template
    if with_solvent:
        solvent = OBC2(topology)
        solvent.to_file(output_handler.get_solvent_template_path())

    if conformation_path is not None:
        conformations = BCEConformations(topology, conformation_path)
        conformations.calculate()
        conformations.save(output_handler.get_conformation_library_path())

    log.info(' - All files were generated successfully:')
    if conformation_path is None:
        log.info('   - {}'.format(output_handler.get_rotamer_library_path()))
    log.info('   - {}'.format(output_handler.get_impact_template_path()))
    if conformation_path is not None:
        log.info('   - {}'.format(output_handler.get_conformation_library_path()))
    if with_solvent:
        log.info('   - {}'.format(output_handler.get_solvent_template_path()))

    log.info('-' * 60)


def main(args):
    """
    It reads the command-line arguments and runs peleffy.

    Parameters
    ----------
    args : argparse.Namespace
        It contains the command-line arguments that are supplied by the user

    Examples
    --------

    From the command-line:

    >>> python main.py molecule.pdb -f openff_unconstrained-1.2.0.offxml
        -r 30 -o output_path/ --with_solvent --as_datalocal -c gasteiger

    """

    exclude_terminal_rotamers = not args.include_terminal_rotamers

    # Supress OpenForceField toolkit warnings
    import logging
    logging.getLogger().setLevel(logging.ERROR)

    # Set peleffy logger to the corresponding level
    logger = Logger()
    if args.silent:
        logger.set_level('CRITICAL')
    elif args.debug:
        logger.set_level('DEBUG')
    else:
        logger.set_level('INFO')

    run_peleffy(pdb_file=args.pdb_file,
                forcefield_name=args.forcefield,
                resolution=args.resolution,
                charge_method=args.charge_method,
                exclude_terminal_rotamers=exclude_terminal_rotamers,
                output=args.output,
                with_solvent=args.with_solvent,
                as_datalocal=args.as_datalocal,
                chain=args.chain,
                conformation_path=args.conformations_info_path,
                charges_from_file=args.charges_from_file)


if __name__ == '__main__':
    import sys
    args = parse_args(sys.argv[1:])
    main(args)
